# -*- coding: utf-8 -*-

"""
Module that provides a variety of overlays for all-sky images.  These overlays
include:
  * the locations and names of sources, 
  * the horizon, 
  * a graticle showing lines of constant RA and dec., and
  * a graticle showing lines of constant azimuth and elevation.

All of the functions in this module accept a matplotlib axes instances that 
is used for plotting.

.. versionadded:: 1.0.1

.. versionchanged:: 1.1.0
    Added support for overlaying on images with non-zenith phase centers
"""

import aipy
import ephem
import numpy

from lsl import astro

__version__ = "0.3"
__revision__ = "$Rev: 2392 $"
__all__ = ["sources", "horizon", "graticleRADec", "graticleAzEl", 
           "__version__", "__revision__", "__all__"]


def _radec_of(aa, az, alt):
    # az/el -> HA/dec
    HA = numpy.arctan2(numpy.sin(az-numpy.pi), (numpy.cos(az-numpy.pi)*numpy.sin(aa.lat) + numpy.tan(alt)*numpy.cos(aa.lat)))
    dec = numpy.arcsin(numpy.sin(aa.lat)*numpy.sin(alt) - numpy.cos(aa.lat)*numpy.cos(alt)*numpy.cos(az-numpy.pi))
    
    # HA -> RA
    RA = aa.sidereal_time() - HA
    
    # radians -> degrees
    RA = RA * 180.0/numpy.pi
    RA %= 360.0
    dec = dec * 180.0/numpy.pi
    
    # RA/dec -> astro.eqn_posn()
    pos = astro.equ_posn(RA, dec)
    
    # Correct for aberration
    pos2 = astro.get_equ_aber(pos, aa.date+astro.DJD_OFFSET)
    dRA, dDec = pos2.ra - pos.ra, pos2.dec - pos.dec
    pos.ra = (pos.ra - dRA) % 360.0
    pos.ra %= 360.0
    pos.dec = pos.dec - dDec
    
    # Correct for nutation
    pos2 = astro.get_equ_nut(pos, aa.date+astro.DJD_OFFSET)
    dRA, dDec = pos2.ra - pos.ra, pos2.dec - pos.dec
    pos.ra = (pos.ra - dRA) % 360.0
    pos.ra %= 360.0
    pos.dec = pos.dec - dDec
    
    # Precess back to J2000
    pos = astro.get_precession(aa.date+astro.DJD_OFFSET, pos, ephem.J2000+astro.DJD_OFFSET)
    RA, dec = pos.ra, pos.dec
    
    # degrees -> radians
    RA = RA * numpy.pi/180.0
    dec = dec * numpy.pi/180.0
    
    return RA, dec 


def sources(ax, aa, srcs, phaseCenter='z', label=True, marker='x', color='white'):
    """
    For a matplotlib axis instance showing an image of the sky, plot the
    locations of the srcs given in the 'srcs' dictionary.
    """
    
    # Get the phase center
    if phaseCenter is not 'z':
        phaseCenter.compute(aa)
        pcRA, pcDec = phaseCenter.ra, phaseCenter.dec
    else:
        pcRA, pcDec = aa.sidereal_time(), aa.lat
    rot = aipy.coord.eq2top_m(0, pcDec)
        
    # Compute the positions of major sources and label the images
    for name,src in srcs.iteritems():
        src.compute(aa)
        eq = aipy.coord.radec2eq((src.ra-pcRA, src.dec))
        top = numpy.dot(rot, eq)
        junk,alt = aipy.coord.top2azalt(top)
        if alt >= 0:
            ax.plot(top[0], top[1], marker=marker, markerfacecolor='None', markeredgecolor=color, 
                linewidth=10.0, markersize=10)
            if label:
                ax.text(top[0], top[1], name, color=color, size=12)


def horizon(ax, aa, phaseCenter='z', color='white'):
    """
    For a matplotlib axis instance showing an image of the sky, plot the horizon.
    
    .. versionchanged:: 1.1.0
        Added a new argument for the AntennaArray instance to provide a 
        uniform default call for all functions.
    """
        
    # Get the phase center
    if phaseCenter is not 'z':
        phaseCenter.compute(aa)
        pcRA, pcDec = phaseCenter.ra, phaseCenter.dec
    else:
        pcRA, pcDec = aa.sidereal_time(), aa.lat
    rot = aipy.coord.eq2top_m(0, pcDec)
    
    # Add in the horizon
    x = numpy.zeros(361) + numpy.nan
    y = numpy.zeros(361) + numpy.nan
    for i in xrange(361):
        ra, dec = _radec_of(aa, i*numpy.pi/180.0, 0.0)
        eq = aipy.coord.radec2eq((ra-pcRA,dec))
        top = numpy.dot(rot, eq)
        junk,alt = aipy.coord.top2azalt(top)
        if alt >= -1e-5:
            x[i] = top[0]
            y[i] = top[1]
    ax.plot(x, y, color=color)


def graticleRADec(ax, aa, phaseCenter='z', label=True, color='white'):
    """
    For a matplotlib axis instance showing an image of the sky, plot lines of
    constant declinate and RA.  Declinations are spaced at 20 degree intervals
    and RAs are spaced at 2 hour intervals.
    """
    
    # Get the phase center
    if phaseCenter is not 'z':
        phaseCenter.compute(aa)
        pcRA, pcDec = phaseCenter.ra, phaseCenter.dec
    else:
        pcRA, pcDec = aa.sidereal_time(), aa.lat
    rot = aipy.coord.eq2top_m(0, pcDec)
    
    # Lines of constant declination first
    decs = range(-80, 90, 20)
    ras = numpy.linspace(0, 360, 800)
    
    x = numpy.zeros(ras.size) + numpy.nan
    y = numpy.zeros(ras.size) + numpy.nan
    for dec in decs:
        x *= numpy.nan
        y *= numpy.nan
        
        # Loop over RA to compute the topocentric coordinates (used by the image) for
        # the lines.  Also, figure out the elevation for each point on the line so
        # we can mask those below the horizon
        for i,ra in enumerate(ras):
            eq = aipy.coord.radec2eq((ra*numpy.pi/180-pcRA, dec*numpy.pi/180))
            top = numpy.dot(rot, eq)
            junk,alt = aipy.coord.top2azalt(top)
            if alt >= -1e-5:
                x[i] = top[0]
                y[i] = top[1]
                
        ax.plot(x, y, color=color, alpha=0.75)
        
        eq = aipy.coord.radec2eq((pcRA-pcRA, (dec+5)*numpy.pi/180))
        top = numpy.dot(rot, eq)
        az,alt = aipy.coord.top2azalt(top)
        if alt > 15*numpy.pi/180 and label:
            ax.text(top[0], top[1], '%+i$^\circ$' % dec, color=color)
            
    # Lines of constant RA			
    decs = numpy.linspace(-80, 80, 400)
    ras = range(0, 360, 30)
    
    x = numpy.zeros(decs.size) + numpy.nan
    y = numpy.zeros(decs.size) + numpy.nan
    for ra in ras:
        x *= numpy.nan
        y *= numpy.nan
        
        # Loop over dec to compute the topocentric coordinates (used by the image) for
        # the lines.  Also, figure out the elevation for each point on the line so
        # we can mask those below the horizon
        for i,dec in enumerate(decs):
            eq = aipy.coord.radec2eq((ra*numpy.pi/180-pcRA, dec*numpy.pi/180))
            top = numpy.dot(rot, eq)
            junk,alt = aipy.coord.top2azalt(top)
            if alt >= -1e-5:
                x[i] = top[0]
                y[i] = top[1]
                
        ax.plot(x, y, color=color, alpha=0.75)
        
        eq = aipy.coord.radec2eq((ra*numpy.pi/180-pcRA, 0))
        top = numpy.dot(rot, eq)
        az,alt = aipy.coord.top2azalt(top)
        if alt > 20*numpy.pi/180 and label:
            ax.text(top[0], top[1], '%i$^h$' % (ra/15,), color=color)


def graticleAzEl(ax, aa, phaseCenter='z', label=True, color='white'):
    """
    For a matplotlib axis instance showing an image of the sky, plot lines of
    constant azimuth and elevation.  Elevations are spaced at 20 degree intervals
    and azimuths are spaced at 45 degree intervals
    """
    
    # Get the phase center
    if phaseCenter is not 'z':
        phaseCenter.compute(aa)
        pcRA, pcDec = phaseCenter.ra, phaseCenter.dec
    else:
        pcRA, pcDec = aa.sidereal_time(), aa.lat
    rot = aipy.coord.eq2top_m(0, pcDec)
    
    # Lines of constant elevation
    els = range(0, 90, 20)
    
    x = numpy.zeros(361) + numpy.nan
    y = numpy.zeros(361) + numpy.nan
    for el in els:
        x *= numpy.nan
        y *= numpy.nan
        
        for i in xrange(361):
            ra, dec = _radec_of(aa, i*numpy.pi/180.0, el*numpy.pi/180.0)
            eq = aipy.coord.radec2eq((ra-pcRA,dec))
            top = numpy.dot(rot, eq)
            junk,alt = aipy.coord.top2azalt(top)
            if alt >= -1e-5:
                x[i] = top[0]
                y[i] = top[1]
                
        ax.plot(x, y, color=color)
        
        if el > 0 or phaseCenter is not 'z':
            valid = numpy.where( numpy.isfinite(x) & numpy.isfinite(y) )[0]
            pos = valid.size / 2 - valid.size / 5
            if valid.size > 10:
                ax.text(x[valid[pos]], y[valid[pos]], '%i$^\circ$' % el, color=color)
            
    # Lines of constant azimuth
    azs = range(0, 360, 45)
    
    x = numpy.zeros(81) + numpy.nan
    y = numpy.zeros(81) + numpy.nan
    for az in azs:
        x *= numpy.nan
        y *= numpy.nan
        
        for i in xrange(81):
            ra, dec = _radec_of(aa, az*numpy.pi/180.0, i*numpy.pi/180.0)
            eq = aipy.coord.radec2eq((ra-pcRA,dec))
            top = numpy.dot(rot, eq)
            junk,alt = aipy.coord.top2azalt(top)
            if alt >= -1e-5:
                x[i] = top[0]
                y[i] = top[1]
                
        ax.plot(x, y, color=color)
        
        valid = numpy.where( numpy.isfinite(x) & numpy.isfinite(y) )[0]
        pos = valid.size / 2 - valid.size / 5
        if valid.size > 10:
            ax.text(x[valid[pos]], y[valid[pos]], '%i$^\circ$' % az, color=color)



"""
Module to help display color on the command line of ANSI-compliant
termainals.

..versionadded:: 1.2.1
"""

import re

__version__ = '0.1'
__revision__ = '$Rev$'
__all__ = ['colorfy', '__version__', '__revision__', '__all__']


# Dictionary of ANSI color codes
_colors = {'black':       u'\u001b[30m', 
        'red':         u'\u001b[31m', 
        'green':       u'\u001b[32m', 
        'yellow':      u'\u001b[33m', 
        'blue':        u'\u001b[34m', 
        'magenta':     u'\u001b[35m', 
        'cyan':        u'\u001b[36m', 
        'white':       u'\u001b[37m', 
        'bkg-black':   u'\u001b[40m', 
        'bkg-red':     u'\u001b[41m', 
        'bkg-green':   u'\u001b[42m', 
        'bkg-yellow':  u'\u001b[43m', 
        'bkg-blue':    u'\u001b[44m', 
        'bkg-magenta': u'\u001b[45m', 
        'bkg-cyan':    u'\u001b[46m', 
        'bkg-white':   u'\u001b[47m', 
        'reset':       u'\u001b[0m', 
        'bold':        u'\u001b[1m', 
        'underline':   u'\u001b[4m', 
        'blink':       u'\u001b[5m',
        'reverse':     u'\u001b[7m'}


def _label(text):
    """
    Private function to tag a ANSI tagged string and split it into various
    tag closure levels.  Returns the level-tagged string.
    """
    
    # Setup the control variables for position in the string and nesting depth
    level = 0
    pos = 0
    
    # Go, until we run out of string
    while pos < len(text):
        ## Split the string at the current position so that we can search for
        ## '{{' and '}}'
        pre  = text[:pos]
        tag  = text[pos:pos+2]
        post = text[pos+2:]
        
        ## Search for tag opening/closing - For the opening we need to make sure 
        ## there is a '%' immediately afterwards.  For closing we need to make 
        ## sure we actually have a valid level to close.
        if tag == '{{':
            if post[0] == '%':
                tag = '{{'+str(level)+tag
                level += 1
                pos = pos + 4
        if tag == '}}':
            level -=1
            if level >= 0:
                tag = tag+str(level)+'}}'
                pos = pos + 4
        text = pre+tag+post
        pos += 1
        
    # Done
    return text


def colorfy(text):
    """
    Given a string encoded with color and/or style information, return a string with
    embedded ANSI terminal codes.  For example:
    {{%red text}} {{%underline underline}}
    will return the text 'text underline' with 'text' displayed as red and 'underline'
    underlined.
    
    Valid colors are:
    * black
    * red
    * green
    * yellow
    * blue
    * magenta
    * cyan
    * white
    
    Valid type styles are:
    * bold
    * underline
    * blink
    * reverse
    
    .. note::
        Background colors can be set by appending 'bkg-' to the color name.
    """
    
    # Parse 1 - Convert nested tags into numbered tags
    text = _label(text)
    
    # Parse 2 - Find the maximum nesting depth of the tags
    nestRE = re.compile(r'{{(?P<depth>\d+){{%')
    levels = [int(l, 10) for l in nestRE.findall(text)]
    levels.append(0)
    
    # Parse 3 - Work at each level to convert the tags to ANSI sequences
    for level in xrange(max(levels), -1, -1):
        ## Build the regular expression needed for this depth
        tagRE = re.compile('{{'+str(level)+'{{%(?P<tag>[a-zA-Z-]+) (?P<text>.*?)}}'+str(level)+'}}')
        
        ## Match and replace until there is nothing left to match
        pos = 0
        mtch = tagRE.search(text, pos)
        while mtch is not None and pos < len(text):
            ### Parse the match
            tag, txt = mtch.group('tag'), mtch.group('text')
            tag = tag.lower()
            try:
                tag = _colors[tag]
            except KeyError:
                tag = ''
                
            ### Update the search start position so we can move on down the string
            pos = mtch.start() + 1
            
            ### Substitute, taking into account any higher levels that have already been encoded
            repl = tag+txt
            repl = repl.replace(_colors['reset'], _colors['reset']+tag)
            repl = repl+_colors['reset']
            text = tagRE.sub(repl, text, count=1)
            
            ### Get the next match
            mtch = tagRE.search(text, pos)
            
    # Done
    return text



# -*- coding: utf-8 -*-

# Python3 compatiability
from __future__ import print_function

"""
Module that contains the error classes for the DRX, TBN, and TBW readers.  
These errors are currently meant to deal with file I/O problems.

.. versionchanged::1.1.4
    Removed numpyError and re-enumerated
"""

__version__ = '0.2'
__revision__ = '$Rev: 2454 $'
__all__ = ['baseReaderError', 'eofError', 'syncError', 'notTBNError', 'notTBWError', 'listErrorCodes', 
           'MinErrorNo', 'MaxErrorNo', 
           '__version__', '__revision__', '__all__']


MinErrorNo = 1
MaxErrorNo = 4


class baseReaderError(IOError):
    """
    Base class for file I/O problem during numpy.fromfile calls and out-of-
    sync Mark5C headers.
    """

    def __init__(self, strerror, errno='-1'):
        self.errno = errno
        self.strerror = strerror
        self.filename = None
        self.args = (errno, strerror)
        super(baseReaderError, self).__init__()

    def __str__(self):
        return "%s" % self.strerror


class eofError(baseReaderError):
    """
    Extension to the base class for dealing with EOF errors.  The error code
    is 1.
    """

    def __init__(self):
        self.errno = 1
        self.strerror = 'End of file encountered during filehandle read'
        self.filename = None
        self.args = (self.errno, self.strerror)
        super(eofError, self).__init__(self.strerror, errno=self.errno)


class syncError(baseReaderError):
    """
    Extension to the base class for dealing with Mark 5C header sync word 
    problems.  If the sync word doesn't match what is expected.  The error code 
    is 3.
    """

    def __init__(self, location=None, sync1=None, sync2=None, sync3=None, sync4=None):
        self.errno = 2
        self.strerror = 'Mark 5C sync word differs from expected'
        self.filename = None
        self.args = (self.errno, self.strerror)
        self.location = location
        self.syncWord = (sync1, sync2, sync3, sync4)
        super(syncError, self).__init__(self.strerror, errno=self.errno)

    def __str__(self):
        output = self.strerror
        if self.location is not None:
            output = '%s at byte %i' % (output, self.location)
        if self.syncWord[0] is not None:
            output = '%s: %02X %02X %02X %02X' % (output, self.syncWord[0], self.syncWord[1], self.syncWord[2], self.syncWord[3])
        return output


class notTBNError(baseReaderError):
    """
    Extenstion to the base class for dealing with trying to read in TBW data 
    with a TBN reader.  The error code is 4.
    """

    def __init__(self):
        self.errno = 3
        self.strerror = 'Data appears to be TBW, not TBN as expected'
        self.filename = None
        self.args = (self.errno, self.strerror)
        super(notTBNError, self).__init__(self.strerror, errno=self.errno)


class notTBWError(baseReaderError):
    """
    Extenstion to the base class for dealing with trying to read in TBN data 
    with a TBW reader.  The error code is 5.
    """

    def __init__(self):
        self.errno = 4
        self.strerror = 'Data appears to be TBN, not TBW as expected'
        self.filename = None
        self.args = (self.errno, self.strerror)
        super(notTBWError, self).__init__(self.strerror, errno=self.errno)


def listErrorCodes(errno=None):
    """
    Function to provide a list of errors defined in this file.  It 
    alternatively takes an error code using the 'errno' keyword and returns its
    description.
    """

    if errno is None:
        for i in range(MinErrorNo, (MaxErrorNo+1)):
            listErrorCodes(errno=i)
    else:
        if errno == 1:
            print("1: End of file encountered during filehandle read")
        elif errno == 2:
            print("2: Mark 5C sync word differs from expected")
        elif errno == 3:
            print("3: Data appears to be TBW, not TBN as expected")
        elif errno == 4:
            print("4: Data appears to be TBN, not TBW as expected")
        else:
            print("Unknown error code '%i'" % errno)



# -*- coding: utf-8 -*-

# Python3 compatiability
from __future__ import division
import sys
if sys.version_info > (3,):
    xrange = range
    
"""
Useful math functions for LWA work.
"""

import logging
import math

import numpy
from scipy.special import sph_harm

__version__   = '0.5'
__revision__ = '$Rev: 2469 $'
__all__ = ['regrid', 'downsample', 'smooth', 'cmagnitude', 'cphase', 'cpolar', 'crect', 
           'creal', 'cimag', 'to_dB', 'from_dB', 'ndft',  'savitzky_golay', 'gaussian1d', 
           'gaussian2d', 'gaussparams', 'sphfit', 'sphval', 
           '__version__', '__revision__', '__all__']
__author__    = 'P. S. Ray'
__maintainer__ = 'Jayce Dowell'

_MATHUTIL_LOG = logging.getLogger('mathutil')


def regrid(x, y, newx, allow_extrapolation = False, method = 'spline'):
    """
    Regrid data from x,y onto newx. If allow_extrapolation is True,
    extrapolation is attempted if the method supports it.  Supported
    methods are:
      * linear
      * spline
    Use of this function may require the scipy extension package.
    """
    
    if method == 'linear':
        return _regrid_linear(x, y, newx, allow_extrapolation)
    elif method == 'spline':
        return _regrid_spline(x, y, newx, allow_extrapolation)
    else:
        raise ValueError("interp method '%s' not recognized" % method)


def _regrid_linear(x, y, newx, allow_extrapolation=False):
    """
    Implement regrid() function using linear interpolation.
    """
    
    if allow_extrapolation:
        _MATHUTIL_LOG.warning("allow_extrapolation=True not honored for regrid_linear")
    
    if newx.min() < x.min():
        raise ValueError('x.min(%f) must be smaller than newx.min(%f)' % (x.min(), newx.min()))
    if newx.max() > x.max():
        raise ValueError('x.max(%f) must be larger than newx.max(%f)' % (x.max(), newx.max()))
        
    return numpy.interp(newx, x, y)


def _regrid_spline(x, y, newx, allow_extrapolation=False):
    """
    Implement regrid() function using spline fit and interpolation.
    """
    
    from scipy import interpolate

    if not allow_extrapolation:
        if newx.min() < x.min():
            raise ValueError('x.min(%f) must be smaller than newx.min(%f)' % (x.min(), newx.min()))
        if newx.max() > x.max():
            raise ValueError('x.max(%f) must be larger than newx.max(%f)' % (x.max(), newx.max()))
    spl = interpolate.splrep(x, y)
    newy = interpolate.splev(newx, spl)
    
    return newy


def downsample(vector, factor, rescale=True):
    """
    Downsample (i.e. co-add consecutive numbers) a vector by an integer 
    factor.  Trims the input timeseries to be a multiple of the downsample 
    factor, if needed.  If rescale == True, then divides each sum by factor 
    to produce a mean value, otherwise just adds the values in the vector.
    """

    if (len(vector) % factor):
        _MATHUTIL_LOG.warning("Length of 'vector' is not divisible by 'factor'=%d, clipping!", factor)
        newlen = (len(vector)//factor)*factor
        _MATHUTIL_LOG.debug("Oldlen %d, newlen %d", len(vector), newlen)
        vector = vector[:newlen]
    if rescale:
        newvector = numpy.reshape(vector, (len(vector)//factor, factor))/float(factor)
    else:
        newvector = numpy.reshape(vector, (len(vector)//factor, factor))
        
    return numpy.add.reduce(newvector, 1)


def smooth(x, window_len=10, window='hanning'):
    """
    Smooth the data using a window with requested size.  Stolen from SciPy 
    Cookbook at http://www.scipy.org/Cookbook/SignalSmooth
    
    This method is based on the convolution of a scaled window with the signal.
    The signal is prepared by introducing reflected copies of the signal 
    (with the window size) in both ends so that transient parts are minimized
    in the begining and end part of the output signal.
    
    Input:
      * x: the input signal 
      * window_len: the dimension of the smoothing window
      * window: the type of window from 'flat', 'hanning', 'hamming', 
                'bartlett', 'blackman' flat window will produce a moving 
                average smoothing.

    Output:
      * the smoothed signal
      
    Example:
        >>> from numpy import *
        >>> t=linspace(-2,2,0.1)
        >>> x=sin(t)+randn(len(t))*0.1
        >>> y=smooth(x)
    
    .. seealso:: 
        numpy.hanning, numpy.hamming, numpy.bartlett, numpy.blackman, numpy.convolve
        scipy.signal.lfilter
    
    TODO: the window parameter could be the window itself if an array instead of a string
    """
    
    if x.ndim != 1:
        raise ValueError("smooth only accepts 1 dimension arrays.")
        
    if x.size < window_len:
        raise ValueError("Input vector needs to be bigger than window size.")
        
    if window_len < 3:
        return x
        
    if not window in ['flat', 'hanning', 'hamming', 'bartlett', 'blackman']:
        raise ValueError("Window is on of 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'")
        
    s = numpy.r_[2*x[0]-x[window_len:1:-1],x,2*x[-1]-x[-1:-window_len:-1]]
    #print(len(s))
    if window == 'flat': #moving average
        w = numpy.ones(window_len, dtype=s.dtype)
    else:
        w = eval('numpy.'+window+'(window_len)')
        
    y = numpy.convolve(w/w.sum(), s, mode='same')
    return y[window_len-1:-window_len+1]


def cmagnitude(cmplx):
    """
    Return the polar magnitudes of complex values.
    """
    
    return abs(cmplx)


def cphase(cmplx):
    """
    Return the polar phases of complex values as radians.
    """
    
    return numpy.angle(cmplx)


def cpolar(cmplx):
    """
    Return the polar (magnitude, phase) representation of complex
    values (real, imaginary).  The return value is an array of shape (N,2),
    where N is the length of the cmplx input array.
    """
    
    if isinstance(cmplx, (numpy.ndarray, list, tuple)):
        return numpy.array(list(zip(cmagnitude(cmplx), cphase(cmplx))))
    else:
        return (cmagnitude(cmplx), cphase(cmplx))


def creal(cmplx):
    """
    Return the real rectilinear component from complex values
    expressed in polar form (magnitude, phase).
    """
    
    if isinstance(cmplx, numpy.ndarray):
        return (cmplx[...,0] * numpy.cos(cmplx[...,1]))
    else:
        return (cmplx[0] * math.cos(cmplx[1]))


def cimag(cmplx):
    """
    Return the imaginary rectilinear component from complex values
    expressed in polar form (magnitude, phase).
    """
    
    if isinstance(cmplx, numpy.ndarray):
        return (cmplx[...,0] * numpy.sin(cmplx[...,1]))
    else:
        return (cmplx[0] * math.sin(cmplx[1]))


def crect(cmplx):
    """
    Return the rectilinear (real, imaginary) representation of complex
    values (magnitude, phase).
    """
    
    if isinstance(cmplx, numpy.ndarray):
        ret = numpy.empty((len(cmplx),), numpy.complex64 if cmplx.dtype == numpy.float32 else numpy.complex128)
        ret.real = creal(cmplx)
        ret.imag = cimag(cmplx)
        return ret         
    else:
        return complex(creal(cmplx), cimag(cmplx))


def to_dB(factor):
    """
    Convert from linear units to decibels.
    """
    
    return 10.0 * numpy.log10(factor)


def from_dB(dB):
    """
    Convert from decibels to linear units.
    """
    
    return numpy.power(10.0, (dB/10.0))


def ndft(t, x):
    """
    Given a list of times and a list of data values, compute a non-uniform 
    discrete Fourier transform (NDFT) of the data.  Returns a two element 
    tuple of frequency and the complex NDFT result.
    """
    
    # Setup the output dtype to make sure that we have enough precision
    if x.dtype in (numpy.complex128, numpy.float64):
        dtype = numpy.complex128
    else:
        dtype = numpy.complex64
        
    # Create the output NDFT array and fill it
    out = numpy.zeros(x.shape, dtype=dtype)
    for m in xrange(out.size):
        mPrime = out.size//2 - m
        s = 0.0j
        for n in xrange(out.size):
            s += x[n]*numpy.exp(-2j*numpy.pi*t[n]*mPrime / (t.max() - t.min()))
        out[m] = s
        
    # Create the output frequency array and fill it
    f = numpy.zeros_like(t)
    for m in xrange(out.size):
        mPrime = out.size//2 - m
        f[m] = mPrime / (t.max() - t.min())
        
    # Done
    return f, out


def savitzky_golay(y, window_size, order, deriv=0):
    """
    Smooth (and optionally differentiate) data with a Savitzky-Golay filter.
    The Savitzky-Golay filter removes high frequency noise from data.  It has 
    the advantage of preserving the original shape and features of the signal 
    better than other types of filtering approaches, such as moving averages 
    techhniques.

    Parameters:
      * y: array_like, shape (N,) = the values of the time history of the signal.
      * window_size: int = the length of the window. Must be an odd integer number.
      * order: int = the order of the polynomial used in the filtering.  Must be 
                     less then `window_size` - 1.
      * deriv: int = the order of the derivative to compute (default = 0 means 
                     only smoothing)

    Returns:
      * ys: ndarray, shape (N) = the smoothed signal (or it's n-th derivative).

    Notes:
        The Savitzky-Golay is a type of low-pass filter, particularly suited for 
        smoothing noisy data. The main idea behind this approach is to make for each 
        point a least-square fit with a polynomial of high order over a odd-sized 
        window centered at the point.

    Examples:
        >>> t = numpy.linspace(-4, 4, 500)
        >>> y = numpy.exp( -t**2 ) + numpy.random.normal(0, 0.05, t.shape)
        >>> ysg = savitzky_golay(y, window_size=31, order=4)
        >>> import matplotlib.pyplot as plt
        >>> plt.plot(t, y, label='Noisy signal')
        >>> plt.plot(t, numpy.exp(-t**2), 'k', lw=1.5, label='Original signal')
        >>> plt.plot(t, ysg, 'r', label='Filtered signal')
        >>> plt.legend()
        >>> plt.show()

    References:
    1. A. Savitzky, M. J. E. Golay, Smoothing and Differentiation of Data by 
        Simplified Least Squares Procedures. Analytical Chemistry, 1964, 36 (8), 
        pp 1627-1639.
    2. Numerical Recipes 3rd Edition: The Art of Scientific Computing W.H. Press,
        S.A. Teukolsky, W.T. Vetterling, B.P. Flannery Cambridge University Press 
        ISBN-13: 9780521880688
    """

    try:
        window_size = numpy.abs(numpy.int(window_size))
        order = numpy.abs(numpy.int(order))
    except ValueError as msg:
        raise ValueError("window_size and order have to be of type int")
    if window_size % 2 != 1 or window_size < 1:
        raise TypeError("window_size size must be a positive odd number")
    if window_size < order + 2:
        raise TypeError("window_size is too small for the polynomials order")
    order_range = range(order+1)
    half_window = (window_size -1) // 2
    # precompute coefficients
    b = numpy.mat([[k**i for i in order_range] for k in range(-half_window, half_window+1)])
    m = numpy.linalg.pinv(b).A[deriv]
    # pad the signal at the extremes with
    # values taken from the signal itself
    firstvals = y[0] - numpy.abs( y[1:half_window+1][::-1] - y[0] )
    lastvals = y[-1] + numpy.abs(y[-half_window-1:-1][::-1] - y[-1])
    y = numpy.concatenate((firstvals, y, lastvals))
    return numpy.convolve( m, y, mode='valid')


def gaussian1d(height, center, width):
    """
    Return a function that generates a 1-D gaussian with the specified
    height, mean, and standard deviation.  
    
    Example:
    >>> height = 1
    >>> center = 5.0
    >>> width = 2.1
    >>> gauFnc = guassian1d(height, center, width)
    >>> value = gauFnc(numpy.arange(0, 100))
    
    Based on: http://code.google.com/p/agpy/source/browse/trunk/agpy/gaussfitter.py
    """
    
    width = float(width)
    return lambda x: height*numpy.exp(-(center-x)**2/2.0/width**2)


def gaussian2d(height, centerX, centerY, widthMaj, widthMin, angle=0.0):
    """
    Return a function that generates a 2-D gaussian with the specified 
    height, mean (for both X and Y), standard deviation (for both major and 
    minor axes), and rotation angle from the X axis in degrees.

    Based on: http://code.google.com/p/agpy/source/browse/trunk/agpy/gaussfitter.py
    """

    widthMaj = float(widthMaj)
    widthMin = float(widthMin)
    pa = angle*numpy.pi/180.0
    return lambda x,y: height*numpy.exp(-((((x-centerX)*numpy.cos(pa)+(y-centerY)*numpy.sin(pa))/widthMaj)**2 + ((-(x-centerX)*numpy.sin(pa)+(y-centerY)*numpy.cos(pa))/widthMin)**2)/2.0)


def gaussparams(data, x=None, y=None):
    """
    Estimate the parameters (height, center, width) for a gaussian.  The 
    return order is:
      1-D: height, center, width
      2-D: height, center x, center y, width x, width y, position angle
    
    .. note:: 
        The 2-D fits always return a position angle of zero since the
        routine decomposes the process into two 1-D fits.
    """
    
    total = data.sum()
    height = data.max()
    
    if len(data.shape) == 2:
        # 2-D Data
        if x is None or y is None:
            x, y = numpy.indices(data.shape)
            
        # Break the problem down into two 1-D filts
        profileX = data.sum(axis=1)
        profileX *= height/profileX.max()
        profileY = data.sum(axis=0)
        profileY *= height/profileY.max()
        
        hx, centerX, widthX = gaussparams(profileX, x[:,0])
        hy, centerY, widthY = gaussparams(profileY, y[0,:])
        return height, centerX, centerY, widthX, widthY, 0.0
        
    else:
        # 1-D Data
        if x is None:
            x = numpy.arange(data.size)
        center = (x*data).sum() / total
        width = numpy.sqrt(abs(numpy.sum((x-center)**2*data)/total))
        return height, center, width


def sphfit(az, alt, data, lmax=5, degrees=False, realOnly=False):
    """
    Decompose a spherical or semi-spherical data set into spherical harmonics.  

    Inputs:
      * az: 2-D numpy array of azimuth coordinates in radians or degrees if the 
            `degrees` keyword is set
      * alt: 2-D numpy array of altitude coordinates in radian or degrees if the 
             `degrees` keyword is set
      * data: 2-D numpy array of the data to be fit.  If the data array is purely
             real, then the `realOnly` keyword can be set which speeds up the 
             decomposition
      * lmax: integer setting the maximum order harmonic to fit

    Keywords:
      * degrees: boolean of whether or not the input azimuth and altitude coordinates
         are in degrees or not
      * realOnly: boolean of whether or not the input data is purely real or not.  If
        the data are real, only coefficients for modes >=0 are computed.

    Returned is a 1-D complex numpy array with the spherical harmonic coefficients 
    packed packed in order of increasing harmonic order and increasing mode, i.e.,
    (0,0), (1,-1), (1,0), (1,1), (2,-2), etc.  If the `realOnly` keyword has been 
    set, the negative coefficients for the negative modes are excluded from the 
    output array.
    
    .. note::
        sphfit was designed to fit the LWA dipole response pattern as a function of
        azimuth and elevation.  Elevation angles are mapped to theta angles by adding
        pi/2 so that an elevation of 90 degrees corresponds to a theta of 180 degrees.
        To fit in terms of spherical coordianates, subtract pi/2 from the theta values
        before running.
    """
    
    if degrees:
        rAz = az*numpy.pi/180.0
        rAlt = alt*numpy.pi/180.0
    else:
        rAz = 1.0*az
        rAlt = 1.0*alt
    rAlt += numpy.pi/2
    sinAlt = numpy.sin(rAlt)
    
    if realOnly:
        nTerms = (lmax*(lmax+3)+2)/2
        terms = numpy.zeros(nTerms, dtype=numpy.complex64)
        
        t = 0
        for l in range(lmax+1):
            for m in range(0, l+1):
                Ylm = sph_harm(m, l, rAz, rAlt)
                terms[t] = (data*sinAlt*Ylm.conj()).sum() * (rAz[1,0]-rAz[0,0])*(rAlt[0,1]-rAlt[0,0])
                t += 1
                
    else:
        nTerms = (lmax+1)**2
        terms = numpy.zeros(nTerms, dtype=numpy.complex64)
        
        t = 0
        for l in range(lmax+1):
            for m in range(-l, l+1):
                Ylm = sph_harm(m, l, rAz, rAlt)
                terms[t] = (data*sinAlt*Ylm.conj()).sum()
                t += 1
                
    return terms


def sphval(terms, az, alt, degrees=False, realOnly=False):
    """
    Evaluate a set of spherical harmonic coefficents at a specified set of
    azimuth and altitude coordinates.

    Inputs:
      * terms: 1-D complex numpy array, typically from sphfit
      * az: 2-D numpy array of azimuth coordinates in radians or degrees if the 
            `degrees` keyword is set
      * alt: 2-D numpy array of altitude coordinates in radian or degrees if the 
             `degrees` keyword is set

    Keywords:
      * degrees: boolean of whether or not the input azimuth and altitude coordinates
                 are in degrees or not
      * realOnly: boolean of whether or not the input data is purely real or not.  If
                  the data are real, only coefficients for modes >=0 are computed.

    Returns a 2-D numpy array of the harmoics evalated and summed at the given 
    coordinates.
    
    .. note::
        sphfit was designed to fit the LWA dipole response pattern as a function of
        azimuth and elevation.  Elevation angles are mapped to theta angles by adding
        pi/2 so that an elevation of 90 degrees corresponds to a theta of 180 degrees.
        To spherical harmonics in terms of spherical coordianates, subtract pi/2 from 
        the theta values before running.
    """
    
    if degrees:
        rAz = az*numpy.pi/180.0
        rAlt = alt*numpy.pi/180.0
    else:
        rAz = 1.0*az
        rAlt = 1.0*alt
    rAlt += numpy.pi/2
    
    nTerms = terms.size
    if realOnly:
        lmax = int((numpy.sqrt(1+8*nTerms)-3)/2)

        t = 0
        out = numpy.zeros(az.shape, dtype=numpy.float32)
        for l in range(lmax+1):
            Ylm = sph_harm(0, l, rAz, rAlt)
            out += numpy.real(terms[t]*Ylm)
            t += 1
            for m in range(1, l+1):
                Ylm = sph_harm(m, l, rAz, rAlt)
                out += numpy.real(terms[t]*Ylm)
                out += numpy.real(terms[t]*Ylm.conj()/(-1)**m)
                t += 1
                
    else:
        lmax = int(numpy.sqrt(nTerms)-1)
        
        t = 0
        out = numpy.zeros(az.shape, dtype=numpy.complex64)
        for l in range(lmax+1):
            for m in range(-l, l+1):
                Ylm = sph_harm(m, l, rAz, rAlt)
                out += terms[t]*Ylm
                t += 1
                
    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Example script that reads in TBF data and runs a cross-correlation on it.  
The results are saved in the FITS IDI format.
"""

import os
import sys
import time
import ephem
import numpy
import argparse
from datetime import datetime, timedelta, tzinfo

from lsl import astro
from lsl.reader.ldp import LWASVDataFile
from lsl.common import stations, metabundleADP
from lsl.correlator import uvUtils
from lsl.correlator import fx as fxc
from lsl.correlator._core import XEngine2
from lsl.writer import fitsidi
from lsl.common.constants import c as speedOfLight
from lsl.misc import parser as aph


class UTC(tzinfo):
    """tzinfo object for UTC time."""
    
    def utcoffset(self, dt):
        return timedelta(0)
        
    def tzname(self, dt):
        return "UTC"
        
    def dst(self, dt):
        return timedelta(0)


def processChunk(idf, site, good, filename, intTime=5.0, pols=['xx',], ChunkSize=100):
    """
    Given a lsl.reader.ldp.TBNFile instances and various parameters for the 
    cross-correlation, write cross-correlate the data and save it to a file.
    """
    
    # Get antennas
    antennas = site.getAntennas()
    
    # Get the metadata
    sampleRate = idf.getInfo('sampleRate')
    freq = idf.getInfo('freq1')
    
    # Create the list of good digitizers and a digitizer to Antenna instance mapping.  
    # These are:
    #  toKeep  -> mapping of digitizer number to array location
    #  mapper -> mapping of Antenna instance to array location
    toKeep = [antennas[i].digitizer-1 for i in good]
    mapper = [antennas[i] for i in good]
    
    # Create a list of unqiue stands to know what style of IDI file to create
    stands = set( [antennas[i].stand.id for i in good] )
    
    # Main loop over the input file to read in the data and organize it.  Several control 
    # variables are defined for this:
    #  refTime -> time (in seconds since the UNIX epoch) for the first data set
    #  setTime -> time (in seconds since the UNIX epoch) for the current data set
    refTime = 0.0
    setTime = 0.0
    wallTime = time.time()
    for s in xrange(ChunkSize):
        try:
            readT, t, data = idf.read(intTime)
        except Exception, e:
            print "Error: %s" % str(e)
            continue
            
        ## Prune out what we don't want
        data = data[toKeep,:,:]
        
        ## Split the polarizations
        antennasX, antennasY = [a for i,a in enumerate(antennas) if a.pol == 0 and i in toKeep], [a for i,a in enumerate(antennas) if a.pol == 1 and i in toKeep]
        dataX, dataY = data[0::2,:,:], data[1::2,:,:]
        validX = numpy.ones((dataX.shape[0],dataX.shape[2]), dtype=numpy.uint8)
        validY = numpy.ones((dataY.shape[0],dataY.shape[2]), dtype=numpy.uint8)
        
        ## Apply the cable delays as phase rotations
        for i in xrange(dataX.shape[0]):
            gain = numpy.sqrt( antennasX[i].cable.gain(freq) )
            phaseRot = numpy.exp(2j*numpy.pi*freq*(antennasX[i].cable.delay(freq) \
                                                   -antennasX[i].stand.z/speedOfLight))
            for j in xrange(dataX.shape[2]):
                dataX[i,:,j] *= phaseRot / gain
        for i in xrange(dataY.shape[0]):
            gain = numpy.sqrt( antennasY[i].cable.gain(freq) )
            phaseRot = numpy.exp(2j*numpy.pi*freq*(antennasY[i].cable.delay(freq)\
                                                   -antennasY[i].stand.z/speedOfLight))
            for j in xrange(dataY.shape[2]):
                dataY[i,:,j] *= phaseRot / gain
                
        setTime = t
        if s == 0:
            refTime = setTime
            
        # Setup the set time as a python datetime instance so that it can be easily printed
        setDT = datetime.utcfromtimestamp(setTime)
        setDT.replace(tzinfo=UTC())
        print "Working on set #%i (%.3f seconds after set #1 = %s)" % ((s+1), (setTime-refTime), setDT.strftime("%Y/%m/%d %H:%M:%S.%f"))
        
        # Loop over polarization products
        for pol in pols:
            print "->  %s" % pol
            if pol[0] == 'x':
                a1, d1, v1 = antennasX, dataX, validX
            else:
                a1, d1, v1 = antennasY, dataY, validY
            if pol[1] == 'x':
                a2, d2, v2 = antennasX, dataX, validX
            else:
                a2, d2, v2 = antennasY, dataY, validY
                
            ## Get the baselines
            baselines = uvUtils.getBaselines(a1, antennas2=a2, IncludeAuto=True, Indicies=True)
            blList = []
            for bl in xrange(len(baselines)):
                blList.append( (a1[baselines[bl][0]], a2[baselines[bl][1]]) )
                
            ## Run the cross multiply and accumulate
            vis = XEngine2(d1, d2, v1, v2)
            
            # Select the right range of channels to save
            toUse = numpy.where( (freq>5.0e6) & (freq<93.0e6) )
            toUse = toUse[0]
            
            # If we are in the first polarazation product of the first iteration,  setup
            # the FITS IDI file.
            if s  == 0 and pol == pols[0]:
                pol1, pol2 = fxc.pol2pol(pol)
                
                if len(stands) > 255:
                    fits = fitsidi.ExtendedIDI(filename, refTime=refTime)
                else:
                    fits = fitsidi.IDI(filename, refTime=refTime)
                fits.setStokes(pols)
                fits.setFrequency(freq[toUse])
                fits.setGeometry(site, [a for a in mapper if a.pol == pol1])
                
            # Convert the setTime to a MJD and save the visibilities to the FITS IDI file
            obsTime = astro.unix_to_taimjd(setTime)
            fits.addDataSet(obsTime, readT, blList, vis[:,toUse], pol=pol)
        print "->  Cummulative Wall Time: %.3f s (%.3f s per integration)" % ((time.time()-wallTime), (time.time()-wallTime)/(s+1))
        
    # Cleanup after everything is done
    fits.write()
    fits.close()
    del(fits)
    del(data)
    del(vis)
    return True


def main(args):
    # Parse command line options
    filename = args.filename
    
    # Setup the LWA station information
    if args.metadata is not None:
        try:
            station = stations.parseSSMIF(args.metadata)
        except ValueError:
            station = metabundleADP.getStation(args.metadata, ApplySDM=True)
    else:
        station = stations.lwasv
    antennas = station.getAntennas()
    
    idf = LWASVDataFile(filename)
    
    jd = astro.unix_to_utcjd(idf.getInfo('tStart'))
    date = str(ephem.Date(jd - astro.DJD_OFFSET))
    nFpO = idf.getInfo('nChan') / 12
    sampleRate = idf.getInfo('sampleRate')
    nInts = idf.getInfo('nFrames') / nFpO
    
    # Get valid stands for both polarizations
    goodX = []
    goodY = []
    for i in xrange(len(antennas)):
        ant = antennas[i]
        if ant.getStatus() != 33 and not args.all:
            pass
        else:
            if ant.pol == 0:
                goodX.append(ant)
            else:
                goodY.append(ant)
                
    # Now combine both lists to come up with stands that
    # are in both so we can form the cross-polarization 
    # products if we need to
    good = []
    for antX in goodX:
        for antY in goodY:
            if antX.stand.id == antY.stand.id:
                good.append( antX.digitizer-1 )
                good.append( antY.digitizer-1 )
                
    # Report on the valid stands found.  This is a little verbose,
    # but nice to see.
    print "Found %i good stands to use" % (len(good)/2,)
    for i in good:
        print "%3i, %i" % (antennas[i].stand.id, antennas[i].pol)
        
    # Number of frames to read in at once and average
    nFrames = min([int(args.avg_time*sampleRate), nInts])
    args.offset = idf.offset(args.offset)
    nSets = idf.getInfo('nFrames') / nFpO / nFrames
    nSets = nSets - int(args.offset*sampleRate) / nFrames
    
    centralFreq = idf.getInfo('freq1')
    centralFreq = centralFreq[len(centralFreq)/2]
    
    print "Data type:  %s" % type(idf)
    print "Samples per observations: %i" % nFpO
    print "Sampling rate: %i Hz" % sampleRate
    print "Tuning frequency: %.3f Hz" % centralFreq
    print "Captures in file: %i (%.3f s)" % (nInts, nInts / sampleRate)
    print "=="
    print "Station: %s" % station.name
    print "Date observed: %s" % date
    print "Julian day: %.5f" % jd
    print "Offset: %.3f s (%i frames)" % (args.offset, args.offset*sampleRate)
    print "Integration Time: %.3f s" % (nFrames/sampleRate)
    print "Number of integrations in file: %i" % nSets
    
    # Make sure we don't try to do too many sets
    if args.samples > nSets:
        args.samples = nSets
        
    # Loop over junks of 100 integrations to make sure that we don't overflow 
    # the FITS IDI memory buffer
    s = 0
    leftToDo = args.samples
    basename = os.path.split(filename)[1]
    basename, ext = os.path.splitext(basename)
    while leftToDo > 0:
        fitsFilename = "%s.FITS_%i" % (basename, (s+1),)
        
        if leftToDo > 100:
            chunk = 100
        else:
            chunk = leftToDo
            
        processChunk(idf, station, good, fitsFilename, intTime=args.avg_time, 
                     pols=args.products, ChunkSize=chunk)
                    
        s += 1
        leftToDo = leftToDo - chunk
        
    idf.close()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description='cross-correlate data in a TBF file', 
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
        )
    parser.add_argument('filename', type=str, 
                        help='filename to correlate')
    parser.add_argument('-m', '--metadata', type=str, 
                        help='name of SSMIF or metadata tarball file to use for mappings')
    parser.add_argument('-t', '--avg-time', type=aph.positive_float, default=1.0, 
                        help='time window to average visibilities in seconds')
    parser.add_argument('-s', '--samples', type=aph.positive_int, default=1, 
                        help='number of average visibilities to generate')
    parser.add_argument('-o', '--offset', type=aph.positive_or_zero_float, default=0.0, 
                        help='offset into the file before starting correlation in seconds')
    parser.add_argument('-q', '--quiet', dest='verbose', action='store_false', 
                        help='run %(prog)s in silent mode')
    parser.add_argument('-a', '--all', action='store_true', 
                        help='correlated all dipoles regardless of their status')
    pgroup = parser.add_mutually_exclusive_group(required=True)
    pgroup.add_argument('-x', '--xx', dest='products', action='store_const', const=['xx',], 
                        help='compute only the XX polarization product')
    pgroup.add_argument('-y', '--yy', dest='products', action='store_const', const=['yy',], 
                        help='compute only the YY polarization product')
    pgroup.add_argument('-2', '--two-products', dest='products', action='store_const', const=['xx','yy'], 
                        help='compute only the XX and YY polarization products')
    pgroup.add_argument('-4', '--four-products', dest='products', action='store_const', const=['xx','yy','xy','yx'], 
                        help='compute the XX, XY, YX, and YY polarization products')
    args = parser.parse_args()
    main(args)
    



# -*- coding: utf-8 -*-
#
# Supplemental configuration file for the LWA Software Library documentation 
# build configuration file (conf.py in this directory).  This file should set 
# the version (long and short), year, and release notes information/files.

import sys, os, re


def version2version(version):
	try:
		version, junk = version.split('-', 1)
	except ValueError:
		pass
	parts = version.split('.')
	
	value = 0.0
	for i,p in enumerate(parts):
		value += float(p) / 10**i
		
	return value


def isDevelopment():
	fh = open(os.path.join('..', '..', 'setup.cfg'))
	
	devel = False
	for line in fh:
		if line.find('tag_build =') == 0:
			keyword, value = line.split(' = ')
			if len(value) > 3:
				devel = True
			break
	fh.close()
	
	return devel


def getLongVersion():
	fh = open(os.path.join('..', '..', 'VERSION'))
	longVersion = fh.read().strip()
	fh.close()
	
	if isDevelopment():
		longVersion = "%s-dev" % longVersion
		
	return longVersion


def getShortVersion():
	longVersion = getLongVersion()
	shortVersion = '.'.join(longVersion.split('.')[0:2])
	
	return shortVersion


def updateReleaseNotes():
	refVersion = version2version(getShortVersion())
	inOlder = False
	
	try:
		fh = open(os.path.join('..', '..', 'CHANGELOG'))
		
		versionRE = re.compile(r'^\d+\.\d+(\.\d+)?')
		
		output = "Release Notes\n=============\n\n"
		for line in fh:
			if len(line) == 0:
				continue
				
			mtch = versionRE.match(line)
			if mtch is not None:
				lineChar = '-'
				if version2version(line) < refVersion:
					lineChar = '+'
					if not inOlder:
						output += "Older Versions\n--------------\n"
						inOlder = True
				output += "Version %s%s\n" % (line, lineChar*(len(line)+8))
			else:
				output += "%s" % line
				
		fh.close()
		
		fh = open('rnotes.rst', 'w')
		fh.write(output)
		fh.close()
	except:
		return False
	
	return True
		


#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Script to gather information about the Python interpreter, modules, 
C libraries, numpy installation, and LSL installation to help with 
debugging and install issues.
"""

import os
import re
import sys
import argparse
import platform
import warnings
import subprocess
try:
    import cStringIO as StringIO
except ImportError:
    import StringIO


def main(args):
    #
    # Clean the path
    #
    sys.path = sys.path[1:]

    #
    # Python interpreter
    #
    print "Executable path: %s" % sys.executable
    print "Platform: %s" % sys.platform
    print "Version: %s" % sys.version
    print "API: %s" % sys.api_version
    print "Bits: %s\nLinkage: %s" % platform.architecture()
    print " "
    
    #
    # Python Module Check
    #
    
    ## Required
    for mod in ('numpy', 'scipy', 'pyfits', 'ephem', 'aipy', 'pytz'):
        try:
            exec "import %s" % mod
        except ImportError, e:
            if (str(e)).find('not found') != -1:
                print  "%s: not found" % mod
            else:
                print "%s: WARNING import error '%s'" % (mod, str(e))
        else:
            try:
                version = eval("%s.version.version" % mod)
            except AttributeError:
                try:
                    version = eval("%s.__version__" % mod)
                except AttributeError:
                    try:	
                        versionRE = re.compile(r'%s-(?P<version>[\d\.]+)-py.*' % mod)
                        mtch = versionRE.search(eval("%s.__file__" % mod))
                        version = mtch.group('version')
                    except:
                        version = "unknown"
            print "%s:  version %s" % (mod, version)
            
    ## Optional
    for mod in ('matplotlib', 'h5py', 'psrfits_utils'):
        try:
            exec "import %s" % mod
        except ImportError, e:
            if (str(e)).find('not found') != -1:
                print  "%s: not found" % mod
            #else:
                #print "%s: WARNING import error '%s'" % (mod, str(e))
        else:
            try:
                version = eval("%s.version.version" % mod)
            except AttributeError:
                try:
                    version = eval("%s.__version__" % mod)
                except AttributeError:
                    try:	
                        versionRE = re.compile(r'%s-(?P<version>[\d\.]+)-py.*' % mod)
                        mtch = versionRE.search(eval("%s.__file__" % mod))
                        version = mtch.group('version')
                    except:
                        version = "unknown"
            print "%s:  version %s" % (mod.capitalize(), version)
            
    
    print " "
    
    
    #
    # Library checks
    #
    
    ##  Via 'pkg-config'
    libsFound = []
    for pkgName in ('fftw3f',):
        try:
            pkgQuery = subprocess.Popen(['pkg-config', '--exists', pkgName])
            o, e = pkgQuery.communicate()
            if pkgQuery.returncode == 0:
                pkgQuery = subprocess.Popen(['pkg-config', '--modversion', pkgName], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                o, e = pkgQuery.communicate()
                o = o.replace('\n', '')
                
                print "%s:  version %s" % (pkgName, o)
                libsFound.append( pkgName )
                
        except OSError:
            pass
            
    ## Via 'numpy'
    try:
        from numpy.distutils.system_info import get_info
        
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore",category=DeprecationWarning)
            sys.stdout = StringIO.StringIO()
            atlas_info = get_info('atlas_blas', notfound_action=0)
            sys.stdout = sys.__stdout__
            
        atlas_version = ([v[3:-3] for k,v in atlas_info.get('define_macros',[])
                        if k == 'ATLAS_INFO']+[None])[0]
        if atlas_version:
            libsFound.append( 'atlas' )
            libsFound.append( 'cblas' )
            
    except ImportError:
        pass
        
    ## Via 'ldconfig'
    try:
        p = subprocess.Popen(['ldconfig', '-v'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        o, e = p.communicate()
        o = o.split('\n')
        
        for lib in ('libatlas', 'libcblas', 'libfftw3f', 'libgdbm', 'librt'):
            libBaseName = lib.replace('lib', '')
            if libBaseName in libsFound:
                continue
                
            found = False
            currPath = None
            
            for line in o:
                if len(line) == 0:
                    continue
                elif line[0] != '\t':
                    currPath, junk = line.split(':', 1)
                    continue
                elif line.find(lib) != -1:
                    found = True
                    libFilename = line.split(None, 1)[0]
                    print "%s: found %s" % (lib, os.path.join(currPath, libFilename))
            if not found:
                print "%s: WARNING - not found" % lib
            
    except OSError:
        pass
    print " "
    
    #
    # Compiler check
    #
    import shutil
    import tempfile
    from distutils import sysconfig
    from distutils import ccompiler
    compiler = ccompiler.new_compiler()
    sysconfig.customize_compiler(compiler)
    cc = compiler.compiler
    
    print "Compiler: %s" % cc[0]
    
    tmpdir = tempfile.mkdtemp()
    curdir = os.getcwd()
    os.chdir(tmpdir)
    
    fh = open('test.c', 'w')
    fh.write(r"""#include <omp.h>
#include <stdio.h>
int main() {
#pragma omp parallel
printf("Hello from thread %d, nthreads %d\n", omp_get_thread_num(), omp_get_num_threads());
}
""")
    fh.close()
    
    cmd = cc
    cmd.extend(['-fopenmp', 'test.c', '-o', 'test', '-lgomp'])
    p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    o, e = p.communicate()
    print "Compiler OpenMP Support: %s" % ("Yes" if p.returncode == 0 else "No",)
    if p.returncode != 0:
        o = o.split('\n')[:-1]
        for i in xrange(len(o)):
            o[i] = '  %s' % o[i]
        o = '\n'.join(o)
        e = e.split('\n')[:-1]
        for i in xrange(len(e)):
            e[i] = '  %s' % e[i]
        e = '\n'.join(e)
        
        print "Compiler OpenMP Test Command:"
        print "  %s" % ' '.join(cmd)
        print "Compiler OpenMP Test Output:"
        print o
        print "Compiler OpenMP Test Errors:"
        print e
        
    os.chdir(curdir)
    shutil.rmtree(tmpdir)
    
    p = subprocess.Popen([cc[0], '-v'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    o, e = p.communicate()
    e = e.split('\n')[:-1]
    for i in xrange(len(e)):
        e[i] = '  %s' % e[i]
    e = '\n'.join(e)
    print "Compiler Version:"
    print e
    print " "
    
    #
    # Numpy
    #
    try:
        import numpy
        nfp,junk = os.path.split(numpy.__file__)
        nfp = os.path.join(nfp, 'core', 'umath.so')
        nfp = os.path.realpath(nfp)

        p = subprocess.Popen(['file', nfp], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        o, e = p.communicate()
        junk, numpyLinkage = o.split(None, 1)
        numpyLinkage = numpyLinkage.replace('\n', '')

        nfp, junk = os.path.split(numpy.__file__)
        print "Numpy Path: %s" % nfp
        print "Numpy Version: %s" % numpy.version.version
        print "Numpy Linkage: %s" % numpyLinkage
    except ImportError, e:
        print "Numpy Import Error: %s" % str(e)
    print " "
    
    #
    # LSL
    #
    try:
        import lsl, lsl.version
        lfp,junk = os.path.split(lsl.__file__)
        lfp = os.path.join(lfp, 'correlator', '_core.so')
        lfp = os.path.realpath(lfp)

        p = subprocess.Popen(['file', lfp], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        o, e = p.communicate()
        junk, lslLinkage = o.split(None, 1)
        lslLinkage = lslLinkage.replace('\n', '')

        lfp, junk = os.path.split(lsl.__file__)
        print "LSL Path: %s" % lfp
        print "LSL Version: %s" % lsl.version.version
        print "LSL Linkage: %s" % lslLinkage
    except ImportError, e:
        print "LSL Import Error: %s" % str(e)
    print " "


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description='gather information about the Python interpreter, modules, C libraries, numpy installation, and LSL installation', 
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    args = parser.parse_args()
    main(args)




# -*- coding: utf-8 -*-

# Python3 compatiability
import sys
if sys.version_info > (3,):
    xrange = range
    
"""
Python module to handle the channelization and cross-correlation of TBW and
TBN data.  The main python functions in this module are:
  * calcSpectra - calculate power spectra for a collection of signals
  * FXCorrelator - calculate cross power spectra for a collection of signals
  * FXStokes - calculate Stokes cross power spectra for a collection of signals
               both of which have been deprecated in favor of the new C extension
               based routines listed below.

The main python/C extension functions in this module are:
  * SpecMaster - similar to calcSpectra but uses the _spec module for all 
                 computations and does not support automatic sub-integration
  * SpecMasterP - SpecMaster with a 64-tap uniform DFT filter bank
  * StokesMaster - similar to SpecMaster but computes all four Stokes parameters
  * FXMaster - calculate cross power spectra for a collection of signals

Each function is set up to process the signals in parallel using the 
multiprocessing module and accepts a variety of options controlling the processing
of the data, including various window functions and time averaging.

.. versionchanged:: 1.0.1
    Removed SpecMasterP.

.. versionchanged:: 1.0.0
    All of the functions here now return all 'LFFT' channels.
"""

import ephem
import numpy

from lsl.common.constants import c as speedOfLight
from lsl.common import dp as dp_common
from lsl.correlator import uvUtils, _spec, _stokes, _core

__version__ = '1.0'
__revision__ = '$Rev: 2457 $'
__all__ = ['pol2pol', 'noWindow', 'SpecMaster', 'StokesMaster', 'FXMaster', 'FXStokes', 
           '__version__', '__revision__', '__all__']


def pol2pol(pol):
    """
    Convert a polarization string, e.g., XX/XY or RR/LL, to a numeric :class:`lsl.common.stations.Antenna`
    instance polarization.
    """
    
    pol = pol.upper()
    out = []
    for p in pol:
        if p in ('X', 'R'):
            out.append(0)
        elif p in ('Y', 'L'):
            out.append(1)
        else:
            raise RuntimeError("Unknown polarization code '%s'" % pol)
        
    return out


def noWindow(L):
    """
    Default "empty" windowing function for use with the various routines.  This
    function returned a numpy array of '1's of the specified length.
    """

    return numpy.ones(L)


def SpecMaster(signals, LFFT=64, window=noWindow, verbose=False, SampleRate=None, CentralFreq=0.0, ClipLevel=0):
    """
    A more advanced version of calcSpectra that uses the _spec C extension 
    to handle all of the P.S.D. calculations in parallel.  Returns a two-
    element tuple of the frequencies (in Hz) and PSDs in linear power/RBW.
    
    .. note::
        SpecMaster currently average all data given and does not support the
        SampleAverage keyword that calcSpectra does.
    """
    
    # Figure out if we are working with complex (I/Q) data or only real.  This
    # will determine how the FFTs are done since the real data mirrors the pos-
    # itive and negative Fourier frequencies.
    if signals.dtype.kind == 'c':
        lFactor = 1
        doFFTShift = True
        CentralFreq = float(CentralFreq)
    else:
        lFactor = 2
        doFFTShift = False

    # Calculate the frequencies of the FFTs.  We do this for twice the FFT length
    # because the real-valued signal only occupies the positive part of the 
    # frequency space.
    if SampleRate is None:
        SampleRate = dp_common.fS
    freq = numpy.fft.fftfreq(lFactor*LFFT, d=1.0/SampleRate)
    # Deal with TBW and TBN data in the correct way
    if doFFTShift:
        freq += CentralFreq
        freq = numpy.fft.fftshift(freq)
    freq = freq[:LFFT]
    
    if window is noWindow:
        window = None
        
    if signals.dtype.kind == 'c':
        output = _spec.FPSDC3(signals, LFFT=LFFT, Overlap=1, ClipLevel=ClipLevel, window=window)
    else:
        output = _spec.FPSDR3(signals, LFFT=LFFT, Overlap=1, ClipLevel=ClipLevel, window=window)
        
    return (freq, output)


def StokesMaster(signals, antennas, LFFT=64, window=noWindow, verbose=False, SampleRate=None, CentralFreq=0.0, ClipLevel=0):
    """
    Similar to SpecMaster, but accepts an array of signals and a list of 
    antennas in order to compute the PSDs for the four Stokes parameters: 
    I, Q, U, and V.  Returns a two-element tuple of the frequencies (in Hz) 
    and PSDs in linear power/RBW.  The PSD are three dimensional with 
    dimensions Stokes parameter (0=I, 1=Q, 2=U, 3=V) by stand by channel).
    """
    
    # Figure out if we are working with complex (I/Q) data or only real.  This
    # will determine how the FFTs are done since the real data mirrors the pos-
    # itive and negative Fourier frequencies.
    if signals.dtype.kind == 'c':
        lFactor = 1
        doFFTShift = True
        CentralFreq = float(CentralFreq)
    else:
        lFactor = 2
        doFFTShift = False
        
    # Match the X and Y stand data
    signalsIndex1 = [i for (i, a) in enumerate(antennas) if a.pol == 0]
    signalsIndex2 = [i for (i, a) in enumerate(antennas) if a.pol == 1]
    if len(signalsIndex1) != len(signalsIndex2):
        raise RuntimeError("Supplied data does not contain an equal number of X and Y signals.")

    # Calculate the frequencies of the FFTs.  We do this for twice the FFT length
    # because the real-valued signal only occupies the positive part of the 
    # frequency space.
    if SampleRate is None:
        SampleRate = dp_common.fS
    freq = numpy.fft.fftfreq(lFactor*LFFT, d=1.0/SampleRate)
    # Deal with TBW and TBN data in the correct way
    if doFFTShift:
        freq += CentralFreq
        freq = numpy.fft.fftshift(freq)
    freq = freq[:LFFT]
    
    if window is noWindow:
        window = None
        
    if signals.dtype.kind == 'c':
        output = _stokes.FPSDC3(signals[signalsIndex1], signals[signalsIndex2], LFFT=LFFT, Overlap=1, ClipLevel=ClipLevel, window=window)
    else:
        output = _stokes.FPSDR3(signals[signalsIndex1], signals[signalsIndex2], LFFT=LFFT, Overlap=1, ClipLevel=ClipLevel, window=window)
        
    return (freq, output)


def FXMaster(signals, antennas, LFFT=64, Overlap=1, IncludeAuto=False, verbose=False, window=noWindow, SampleRate=None, CentralFreq=0.0, Pol='XX', GainCorrect=False, ReturnBaselines=False, ClipLevel=0, phaseCenter='z'):
    """
    A more advanced version of FXCorrelator for TBW and TBN data.  Given an 
    2-D array of signals (stands, time-series) and an array of stands, compute 
    the cross-correlation of the data for all baselines.  Return the frequencies 
    and visibilities as a two-elements tuple.
    
    .. versionchanged:: 0.4.0
        Switched over to passing in Antenna instances generated by the
        :mod:`lsl.common.stations` module instead of a list of stand ID
        numbers.
        
    .. versionchanged:: 1.0.0
        Added a phase-center keyword that accept a two-element tuple of 
        azimuth and elelvation (in degrees) to change where the 
        correlations are phased to
        
    .. versionchanged:: 1.1.0
        Made the 'phaseCenter' keyword more flexible.  It can now be either:
        * 'z' to denote the zenith,
        * a ephem.Body instances which has been computed for the observer, or
        * a two-element tuple of azimuth, elevation in degrees.
    """
    
    # Decode the polarization product into something that we can use to figure 
    # out which antennas to use for the cross-correlation
    pol1, pol2 = pol2pol(Pol)
    
    antennas1 = [a for a in antennas if a.pol == pol1]
    signalsIndex1 = [i for (i, a) in enumerate(antennas) if a.pol == pol1]
    antennas2 = [a for a in antennas if a.pol == pol2]
    signalsIndex2 = [i for (i, a) in enumerate(antennas) if a.pol == pol2]
    
    nStands = len(antennas1)
    baselines = uvUtils.getBaselines(antennas1, antennas2=antennas2, IncludeAuto=IncludeAuto, Indicies=True)
    
    # Figure out if we are working with complex (I/Q) data or only real.  This
    # will determine how the FFTs are done since the real data mirrors the pos-
    # itive and negative Fourier frequencies.
    if signals.dtype.kind == 'c':
        lFactor = 1
        doFFTShift = True
        CentralFreq = float(CentralFreq)
    else:
        lFactor = 2
        doFFTShift = False
        
    if SampleRate is None:
        SampleRate = dp_common.fS
    freq = numpy.fft.fftfreq(lFactor*LFFT, d=1.0/SampleRate)
    if doFFTShift:
        freq += CentralFreq
        freq = numpy.fft.fftshift(freq)
    freq = freq[:LFFT]
    
    # Get the location of the phase center in radians and create a 
    # pointing vector
    if phaseCenter == 'z':
        azPC = 0.0
        elPC = numpy.pi/2.0
    else:
        if isinstance(phaseCenter, ephem.Body):
            azPC = phaseCenter.az * 1.0
            elPC = phaseCenter.alt * 1.0
        else:
            azPC = phaseCenter[0]*numpy.pi/180.0
            elPC = phaseCenter[1]*numpy.pi/180.0
            
    source = numpy.array([numpy.cos(elPC)*numpy.sin(azPC), 
                    numpy.cos(elPC)*numpy.cos(azPC), 
                    numpy.sin(elPC)])
                    
    # Define the cable/signal delay caches to help correlate along and compute 
    # the delays that we need to apply to align the signals
    dlyRef = len(freq)/2
    delays1 = numpy.zeros((nStands,LFFT))
    delays2 = numpy.zeros((nStands,LFFT))
    for i in list(range(nStands)):
        xyz1 = numpy.array([antennas1[i].stand.x, antennas1[i].stand.y, antennas1[i].stand.z])
        xyz2 = numpy.array([antennas2[i].stand.x, antennas2[i].stand.y, antennas2[i].stand.z])
        
        delays1[i,:] = antennas1[i].cable.delay(freq) - numpy.dot(source, xyz1) / speedOfLight
        delays2[i,:] = antennas2[i].cable.delay(freq) - numpy.dot(source, xyz2) / speedOfLight
    if not numpy.isfinite(delays1.max()):
        delays1[numpy.where( ~numpy.isfinite(delays1) )] = delays1[numpy.where( numpy.isfinite(delays1) )].max()
    if not numpy.isfinite(delays2.max()):
        delays2[numpy.where( ~numpy.isfinite(delays2) )] = delays2[numpy.where( numpy.isfinite(delays2) )].max()
    if delays1[:,dlyRef].min() < delays2[:,dlyRef].min():
        minDelay = delays1[:,dlyRef].min()
    else:
        minDelay = delays2[:,dlyRef].min()
    delays1 -= minDelay
    delays2 -= minDelay
    
    if window is noWindow:
        window = None
        
    # F - defaults to running parallel in C via OpenMP
    if signals.dtype.kind == 'c':
        FEngine = _core.FEngineC3
    else:
        FEngine = _core.FEngineR3
    if signals.shape[0] != len(signalsIndex1):
        signalsF1, validF1 = FEngine(signals[signalsIndex1,:], freq, delays1, LFFT=LFFT, Overlap=Overlap, SampleRate=SampleRate, ClipLevel=ClipLevel, window=window)
    else:
        signalsF1, validF1 = FEngine(signals, freq, delays1, LFFT=LFFT, Overlap=Overlap, SampleRate=SampleRate, ClipLevel=ClipLevel, window=window)
        
    if pol2 == pol1:
        signalsF2 = signalsF1
        validF2 = validF1
    else:
        signalsF2, validF2 = FEngine(signals[signalsIndex2,:], freq, delays2, LFFT=LFFT, Overlap=Overlap, SampleRate=SampleRate, ClipLevel=ClipLevel, window=window)
        
    # X
    output = _core.XEngine2(signalsF1, signalsF2, validF1, validF2)
    if not IncludeAuto:
        # Remove auto-correlations from the output of the X engine if we don't 
        # need them.  To do this we need to first build the full list of baselines
        # (including auto-correlations) and then prune that.
        baselinesFull = uvUtils.getBaselines(antennas1, antennas2=antennas2, IncludeAuto=True, Indicies=True)
        fom = numpy.array([a1-a2 for (a1,a2) in baselinesFull])
        nonAuto = numpy.where( fom != 0 )[0]
        output = output[nonAuto,:]
        
    # Apply cable gain corrections (if needed)
    if GainCorrect:
        for bl in xrange(output.shape[0]):
            cableGain1 = antennas1[baselines[bl][0]].cable.gain(freq)
            cableGain2 = antennas2[baselines[bl][1]].cable.gain(freq)
            
            output[bl,:] /= numpy.sqrt(cableGain1*cableGain2)
            
    # Create antenna baseline list (if needed)
    if ReturnBaselines:
        antennaBaselines = []
        for bl in xrange(output.shape[0]):
            antennaBaselines.append( (antennas1[baselines[bl][0]], antennas2[baselines[bl][1]]) )
        returnValues = (antennaBaselines, freq, output)
    else:
        returnValues = (freq, output)

    return returnValues


def FXStokes(signals, antennas, LFFT=64, Overlap=1, IncludeAuto=False, verbose=False, window=noWindow, SampleRate=None, CentralFreq=0.0,  GainCorrect=False, ReturnBaselines=False, ClipLevel=0, phaseCenter='z'):
    """
    A more advanced version of FXCorrelator for TBW and TBN data.  Given an 
    2-D array of signals (stands, time-series) and an array of stands, compute 
    the cross-correlation of the data for all baselines.  Return the frequencies 
    and visibilities as a two-elements tuple.
    
    .. versionchanged:: 1.0.0
        Added a phase-center keyword that accept a two-element tuple of 
        azimuth and elelvation (in degrees) to change where the 
        correlations are phased to
        
    .. versionchanged:: 1.1.0
        Made the 'phaseCenter' keyword more flexible.  It can now be either:
        * 'z' to denote the zenith,
        * a ephem.Body instances which has been computed for the observer, or
        * a two-element tuple of azimuth, elevation in degrees.
    """
    
    # Since we want to compute Stokes parameters, we need both pols
    pol1 = 0
    pol2 = 1
    
    antennas1 = [a for a in antennas if a.pol == pol1]
    signalsIndex1 = [i for (i, a) in enumerate(antennas) if a.pol == pol1]
    antennas2 = [a for a in antennas if a.pol == pol2]
    signalsIndex2 = [i for (i, a) in enumerate(antennas) if a.pol == pol2]
    
    nStands = len(antennas1)
    baselines = uvUtils.getBaselines(antennas1, antennas2=antennas2, IncludeAuto=IncludeAuto, Indicies=True)
    
    # Figure out if we are working with complex (I/Q) data or only real.  This
    # will determine how the FFTs are done since the real data mirrors the pos-
    # itive and negative Fourier frequencies.
    if signals.dtype.kind == 'c':
        lFactor = 1
        doFFTShift = True
        CentralFreq = float(CentralFreq)
    else:
        lFactor = 2
        doFFTShift = False

    if SampleRate is None:
        SampleRate = dp_common.fS
    freq = numpy.fft.fftfreq(lFactor*LFFT, d=1.0/SampleRate)
    if doFFTShift:
        freq += CentralFreq
        freq = numpy.fft.fftshift(freq)
    freq = freq[:LFFT]
    
    # Get the location of the phase center in radians and create a 
    # pointing vector
    if phaseCenter == 'z':
        azPC = 0.0
        elPC = numpy.pi/2.0
    else:
        if isinstance(phaseCenter, ephem.Body):
            azPC = phaseCenter.az * 1.0
            elPC = phaseCenter.alt * 1.0
        else:
            azPC = phaseCenter[0]*numpy.pi/180.0
            elPC = phaseCenter[1]*numpy.pi/180.0
    source = numpy.array([numpy.cos(elPC)*numpy.sin(azPC), 
                    numpy.cos(elPC)*numpy.cos(azPC), 
                    numpy.sin(elPC)])
                    
    # Define the cable/signal delay caches to help correlate along and compute 
    # the delays that we need to apply to align the signals
    dlyRef = len(freq)/2
    delays1 = numpy.zeros((nStands,LFFT))
    delays2 = numpy.zeros((nStands,LFFT))
    for i in list(range(nStands)):
        xyz1 = numpy.array([antennas1[i].stand.x, antennas1[i].stand.y, antennas1[i].stand.z])
        xyz2 = numpy.array([antennas2[i].stand.x, antennas2[i].stand.y, antennas2[i].stand.z])
        
        delays1[i,:] = antennas1[i].cable.delay(freq) - numpy.dot(source, xyz1) / speedOfLight
        delays2[i,:] = antennas2[i].cable.delay(freq) - numpy.dot(source, xyz2) / speedOfLight
    if not numpy.isfinite(delays1.max()):
        delays1[numpy.where( ~numpy.isfinite(delays1) )] = delays1[numpy.where( numpy.isfinite(delays1) )].max()
    if not numpy.isfinite(delays2.max()):
        delays2[numpy.where( ~numpy.isfinite(delays2) )] = delays2[numpy.where( numpy.isfinite(delays2) )].max()
    if delays1[:,dlyRef].min() < delays2[:,dlyRef].min():
        minDelay = delays1[:,dlyRef].min()
    else:
        minDelay = delays2[:,dlyRef].min()
    delays1 -= minDelay
    delays2 -= minDelay
    
    if window is noWindow:
        window = None
        
    # F - defaults to running parallel in C via OpenMP
    if signals.dtype.kind == 'c':
        FEngine = _core.FEngineC3
    else:
        FEngine = _core.FEngineR3
    signalsF1, validF1 = FEngine(signals[signalsIndex1,:], freq, delays1, LFFT=LFFT, Overlap=Overlap, SampleRate=SampleRate, ClipLevel=ClipLevel, window=window)
    
    signalsF2, validF2 = FEngine(signals[signalsIndex2,:], freq, delays2, LFFT=LFFT, Overlap=Overlap, SampleRate=SampleRate, ClipLevel=ClipLevel, window=window)
    
    # X
    output = _stokes.XEngine2(signalsF1, signalsF2, validF1, validF2)
    if not IncludeAuto:
        # Remove auto-correlations from the output of the X engine if we don't 
        # need them.  To do this we need to first build the full list of baselines
        # (including auto-correlations) and then prune that.
        baselinesFull = uvUtils.getBaselines(antennas1, antennas2=antennas2, IncludeAuto=True, Indicies=True)
        fom = numpy.array([a1-a2 for (a1,a2) in baselinesFull])
        nonAuto = numpy.where( fom != 0 )[0]
        output = output[:,nonAuto,:]
        
    # Apply cable gain corrections (if needed)
    if GainCorrect:
        for bl in xrange(output.shape[0]):
            cableGain1 = antennas1[baselines[bl][0]].cable.gain(freq)
            cableGain2 = antennas2[baselines[bl][1]].cable.gain(freq)
            
            output[:,bl,:] /= numpy.sqrt(cableGain1*cableGain2)
            
    # Create antenna baseline list (if needed)
    if ReturnBaselines:
        antennaBaselines = []
        for bl in xrange(output.shape[1]):
            antennaBaselines.append( (antennas1[baselines[bl][0]], antennas2[baselines[bl][1]]) )
        returnValues = (antennaBaselines, freq, output)
    else:
        returnValues = (freq, output)

    return returnValues	



# -*- coding: utf-8 -*-

"""Unit test for lsl.common.mcs"""

import os
import unittest
from datetime import datetime

from lsl.common import mcs


__revision__ = "$Rev: 839 $"
__version__  = "0.1"
__author__    = "Jayce Dowell"


class mcs_tests(unittest.TestCase):
    """A unittest.TestCase collection of unit tests for the lsl.common.mcs
    module."""
    
    def test_datetime(self):
        """Test the datetime to MJD, MPM conversion"""
        
        dt = datetime.strptime("2012-06-15 06:34:09", "%Y-%m-%d %H:%M:%S")
        mjd, mpm = mcs.datetime2mjdmpm(dt)
        
        self.assertEqual(mjd, 56093)
        self.assertEqual(mpm, 23649000)
    
    def test_mjdmpm(self):
        """Test the MJD, MPM to datetime conversion"""
        
        mjd, mpm = 56093, 23649000
        dt = mcs.mjdmpm2datetime(mjd, mpm)
        
        self.assertEqual(dt.strftime("%Y-%m-%d %H:%M:%S"), "2012-06-15 06:34:09")
        
    def test_pointing_correction(self):
        """Test the pointing correction function"""
        
        az = 63.4
        el = 34.2
        
        # No rotation
        theta = 0.0
        phi = 0.0
        psi = 0.0
        azP, elP = mcs.applyPointingCorrection(az, el, theta, phi, psi, degrees=True)
        self.assertAlmostEqual(azP, az, 1)
        self.assertAlmostEqual(elP, el, 1)
        
        # Azimuth only
        theta = 0.0
        phi = 0.0
        psi = 1.0
        azP, elP = mcs.applyPointingCorrection(az, el, theta, phi, psi, degrees=True)
        self.assertAlmostEqual(azP, az-1.0, 1)
        self.assertAlmostEqual(elP, el, 1)
        
        # Something random
        theta = 23.0
        phi = 10.0
        psi = 1.5
        azP, elP = mcs.applyPointingCorrection(az, el, theta, phi, psi, degrees=True)
        self.assertAlmostEqual(azP, 62.40, 1)
        self.assertAlmostEqual(elP, 34.37, 1)

    
class mcs_test_suite(unittest.TestSuite):
    """A unittest.TestSuite class which contains all of the lsl.common.mcs
    module unit tests."""
    
    def __init__(self):
        unittest.TestSuite.__init__(self)
        
        loader = unittest.TestLoader()
        self.addTests(loader.loadTestsFromTestCase(mcs_tests))        
        
        
if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

# Python3 compatiability
from __future__ import division

"""Unit test for the lsl.writer.tsfits modules."""

import os
import time
import unittest
import tempfile
import numpy
import pyfits

from lsl.common.paths import dataBuild as dataPath
from lsl.reader import tbw, tbn, vdif as vrdr, errors
from lsl.writer import vdif


__revision__ = "$Rev: 2579 $"
__version__  = "0.2"
__author__    = "Jayce Dowell"

tbwFile = os.path.join(dataPath, 'tests', 'tbw-test.dat')
tbnFile = os.path.join(dataPath, 'tests', 'tbn-test.dat')


class vdif_tests(unittest.TestCase):
    """A unittest.TestCase collection of unit tests for the lsl.writer.vdif
    module."""

    testPath = None

    def setUp(self):
        """Turn off all numpy warnings and create the temporary file directory."""

        numpy.seterr(all='ignore')
        self.testPath = tempfile.mkdtemp(prefix='test-vdif-', suffix='.tmp')

    def __getTBW(self):
        """Private function to load in the test TBW data and get the frames."""

        fh = open(tbwFile, 'rb')

        # Frames 1 through 8
        frames = []
        for i in range(1,9):
            frames.append(tbw.readFrame(fh))

        fh.close()
        return frames

    def __getTBN(self, vanilla=False):
        """Private function to load in the test TBN data and get the frames.  If 
        the keyword 'vanilla' is set to True, gain, sample rate, and frequency meta-
        data are not added to the frames."""

        fh = open(tbnFile, 'rb')

        # Frames 1 through 8
        frames = []
        for i in range(1,9):
            frames.append(tbn.readFrame(fh))

        if not vanilla:
            # Set some values for the other meta-data
            for frame in frames:
                frame.setSampleRate(100000)

        fh.close()
        return frames

    def test_vdif_real(self):
        """Test writing real data to VDIF format."""

        # Setup the file names
        testFile = os.path.join(self.testPath, 'tbw-test-W.fits')

        # Get some data
        frames = self.__getTBW()

        # Write the data
        fh = open(testFile, 'wb')
        for frame in frames:
            vFrame = vdif.Frame(frame.parseID(), frame.getTime(), bits=8, data=frame.data.xy[0,:].astype(numpy.int8))
            vFrame.writeRawFrame(fh)
        fh.close()

        # Read it back in
        fh = open(testFile, 'rb')
        for tFrame in frames:
            vFrame = vrdr.readFrame(fh)
            self.assertAlmostEqual(vFrame.getTime(), tFrame.getTime(), 6)
            for v,t in zip((vFrame.data.data*256-1)/2, tFrame.data.xy[0,:].astype(numpy.int8)):
                self.assertAlmostEqual(v, t, 6)
                
        fh.close()

    def test_vdif_complex(self):
        """Test writing complex data to VIDF format."""

        # Setup the file names
        testFile = os.path.join(self.testPath, 'tbn-test-W.fits')

        # Get some data
        frames = self.__getTBN()
        
        # Write the data
        fh = open(testFile, 'wb')
        for frame in frames:
            stand, pol = frame.parseID()
            if pol == 1:
                continue
            vFrame = vdif.Frame(stand, frame.getTime(), bits=8, data=frame.data.iq)
            vFrame.writeRawFrame(fh)
        fh.close()

        # Read it back in
        fh = open(testFile, 'rb')
        for tFrame in frames[::2]:
            vFrame = vrdr.readFrame(fh)
            for v,t in zip((vFrame.data.data*256-1-1j)/2, tFrame.data.iq):
                self.assertAlmostEqual(v, t, 6)
                
        fh.close()

    def tearDown(self):
        """Remove the test path directory and its contents"""

        tempFiles = os.listdir(self.testPath)
        for tempFile in tempFiles:
            os.unlink(os.path.join(self.testPath, tempFile))
        os.rmdir(self.testPath)
        self.testPath = None


class  vdif_test_suite(unittest.TestSuite):
    """A unittest.TestSuite class which contains all of the lsl.writer.vdif units 
    tests."""
    
    def __init__(self):
        unittest.TestSuite.__init__(self)
        
        loader = unittest.TestLoader()
        self.addTests(loader.loadTestsFromTestCase(vdif_tests)) 


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

"""Unit test for the lsl.writer.fitsidi modules."""

import os
import time
import unittest
import tempfile
import numpy
import pyfits

from lsl.common import stations as lwa_common
from lsl.correlator import uvUtils
from lsl.writer import fitsidi


__version__  = "0.2"
__revision__ = "$Rev: 2675 $"
__author__   = "Jayce Dowell"


class fitsidi_tests(unittest.TestCase):
    """A unittest.TestCase collection of unit tests for the lsl.writer.fitsidi.idi
    class."""

    testPath = None

    def setUp(self):
        """Turn off all numpy warnings and create the temporary file directory."""

        numpy.seterr(all='ignore')
        self.testPath = tempfile.mkdtemp(prefix='test-fitsidi-', suffix='.tmp')

    def __initData(self):
        """Private function to generate a random set of data for writing a FITS
        IDI file.  The data is returned as a dictionary with keys:
        * freq - frequency array in Hz
        * site - lwa.common.stations object
        * stands - array of stand numbers
        * bl - list of baseline pairs in real stand numbers
        * vis - array of visibility data in baseline x freq format
        """

        # Frequency range
        freq = numpy.arange(0,512)*20e6/512 + 40e6
        # Site and stands
        site = lwa_common.lwa1
        antennas = site.getAntennas()[0:40:2]
        
        # Set baselines and data
        blList = uvUtils.getBaselines(antennas, IncludeAuto=True, Indicies=False)
        visData = numpy.random.rand(len(blList), len(freq))
        visData = visData.astype(numpy.complex64)

        return {'freq': freq, 'site': site, 'antennas': antennas, 'bl': blList, 'vis': visData}

    def test_write_tables(self):
        """Test if the FITS IDI writer writes all of the tables."""

        testTime = time.time()
        testFile = os.path.join(self.testPath, 'idi-test-W.fits')
        
        # Get some data
        data = self.__initData()
        
        # Start the file
        fits = fitsidi.IDI(testFile, refTime=testTime)
        fits.setStokes(['xx'])
        fits.setFrequency(data['freq'])
        fits.setGeometry(data['site'], data['antennas'])
        fits.addComment('This is a comment')
        fits.addHistory('This is history')
        fits.addDataSet(testTime, 6.0, data['bl'], data['vis'])
        fits.write()

        # Open the file and examine
        hdulist = pyfits.open(testFile)
        # Check that all of the extensions are there
        extNames = [hdu.name for hdu in hdulist]
        for ext in ['ARRAY_GEOMETRY', 'FREQUENCY', 'ANTENNA', 'BANDPASS', 'SOURCE', 'UV_DATA']:
            self.assertTrue(ext in extNames)
        # Check the comments and history
        self.assertTrue('This is a comment' in str(hdulist[0].header['COMMENT']).split('\n'))
        self.assertTrue('This is history' in str(hdulist[0].header['HISTORY']).split('\n'))
        
        hdulist.close()
        
    def test_writer_errors(self):
        """Test that common FITS IDI error conditions are caught."""
        
        testTime = time.time()
        testFile = os.path.join(self.testPath, 'idi-test-ERR.fits')
        
        # Get some data
        data = self.__initData()
        
        for i in range(4):
            # Start the file
            fits = fitsidi.IDI(testFile, refTime=testTime, clobber=True)
            if i != 0:
                fits.setStokes(['xx'])
            if i != 1:
                fits.setFrequency(data['freq'])
            if i != 2:
                fits.setGeometry(data['site'], data['antennas'])
            if i != 3:
                fits.addDataSet(testTime, 6.0, data['bl'], data['vis'])
            self.assertRaises(RuntimeError, fits.write)
            
    def test_array_geometry(self):
        """Test the ARRAY_GEOMETRY table."""

        testTime = time.time()
        testFile = os.path.join(self.testPath, 'idi-test-AG.fits')
        
        # Get some data
        data = self.__initData()
        
        # Start the file
        fits = fitsidi.IDI(testFile, refTime=testTime)
        fits.setStokes(['xx'])
        fits.setFrequency(data['freq'])
        fits.setGeometry(data['site'], data['antennas'])
        fits.addDataSet(testTime, 6.0, data['bl'], data['vis'])
        fits.write()

        # Open the file and examine
        hdulist = pyfits.open(testFile)
        ag = hdulist['ARRAY_GEOMETRY'].data
        # Correct number of stands
        self.assertEqual(len(data['antennas']), len(ag.field('NOSTA')))

        # Correct stand names
        names = ['LWA%03i' % ant.stand.id for ant in data['antennas']]
        for name, anname in zip(names, ag.field('ANNAME')):
            self.assertEqual(name, anname)

        hdulist.close()

    def test_frequency(self):
        """Test the FREQUENCY table."""

        testTime = time.time()
        testFile = os.path.join(self.testPath, 'idi-test-FQ.fits')
        
        # Get some data
        data = self.__initData()
        
        # Start the file
        fits = fitsidi.IDI(testFile, refTime=testTime)
        fits.setStokes(['xx'])
        fits.setFrequency(data['freq'])
        fits.setGeometry(data['site'], data['antennas'])
        fits.addDataSet(testTime, 6.0, data['bl'], data['vis'])
        fits.write()

        # Open the file and examine
        hdulist = pyfits.open(testFile)
        fq = hdulist['FREQUENCY'].data
        # Correct number of FREQIDs
        self.assertEqual(len(fq.field('FREQID')), 1)

        # Correct channel width
        self.assertAlmostEqual(fq.field('CH_WIDTH')[0], numpy.abs(data['freq'][1]-data['freq'][0]), 4)

        # Correct bandwidth
        self.assertAlmostEqual(fq.field('TOTAL_BANDWIDTH')[0], numpy.abs(data['freq'][-1]-data['freq'][0]).astype(numpy.float32), 4)

        # Correct sideband
        self.assertEqual(fq.field('SIDEBAND')[0], 1)

        hdulist.close()

    def test_antenna(self):
        """Test the ANTENNA table."""

        testTime = time.time()
        testFile = os.path.join(self.testPath, 'idi-test-AN.fits')
        
        # Get some data
        data = self.__initData()
        
        # Start the file
        fits = fitsidi.IDI(testFile, refTime=testTime)
        fits.setStokes(['xx'])
        fits.setFrequency(data['freq'])
        fits.setGeometry(data['site'], data['antennas'])
        fits.addDataSet(testTime, 6.0, data['bl'], data['vis'])
        fits.write()

        # Open the file and examine
        hdulist = pyfits.open(testFile)
        an = hdulist['ANTENNA'].data
        # Correct number of stands
        self.assertEqual(len(data['antennas']), len(an.field('ANTENNA_NO')))

        # Correct FREQIDs
        for freqid in an.field('FREQID'):
            self.assertEqual(freqid, 1)

        hdulist.close()

    def test_bandpass(self):
        """Test the BANDPASS table."""

        testTime = time.time()
        testFile = os.path.join(self.testPath, 'idi-test-BP.fits')
        
        # Get some data
        data = self.__initData()
        
        # Start the file
        fits = fitsidi.IDI(testFile, refTime=testTime)
        fits.setStokes(['xx'])
        fits.setFrequency(data['freq'])
        fits.setGeometry(data['site'], data['antennas'])
        fits.addDataSet(testTime, 6.0, data['bl'], data['vis'])
        fits.write()

        # Open the file and examine
        hdulist = pyfits.open(testFile)
        bp = hdulist['BANDPASS'].data
        # Correct number of entries
        self.assertEqual(len(data['antennas']), len(bp.field('ANTENNA_NO')))

        # Correct Source ID number
        for src in bp.field('SOURCE_ID'):
            self.assertEqual(src, 0)

        # Correct FREQIDs
        for freqid in bp.field('FREQID'):
            self.assertEqual(freqid, 1)

        hdulist.close()

    def test_source(self):
        """Test the SOURCE table."""

        testTime = time.time()
        testFile = os.path.join(self.testPath, 'idi-test-SO.fits')
        
        # Get some data
        data = self.__initData()
        
        # Start the file
        fits = fitsidi.IDI(testFile, refTime=testTime)
        fits.setStokes(['xx'])
        fits.setFrequency(data['freq'])
        fits.setGeometry(data['site'], data['antennas'])
        fits.addDataSet(testTime, 6.0, data['bl'], data['vis'])
        fits.write()

        # Open the file and examine
        hdulist = pyfits.open(testFile)
        so = hdulist['SOURCE'].data
        # Correct number of entries
        self.assertEqual(len(so.field('SOURCE_ID')), 1)

        # Correct Source ID number
        self.assertEqual(so.field('SOURCE_ID'), 1)

        hdulist.close()

    def test_uvdata(self):
        """Test the UV_DATA table."""

        testTime = time.time()
        testFile = os.path.join(self.testPath, 'idi-test-UV.fits')
        
        # Get some data
        data = self.__initData()
        
        # Start the file
        fits = fitsidi.IDI(testFile, refTime=testTime)
        fits.setStokes(['xx'])
        fits.setFrequency(data['freq'])
        fits.setGeometry(data['site'], data['antennas'])
        fits.addDataSet(testTime, 6.0, data['bl'], data['vis'])
        fits.write()

        # Open the file and examine
        hdulist = pyfits.open(testFile)
        uv = hdulist['UV_DATA'].data

        # Load the mapper
        try:
            mp = hdulist['NOSTA_MAPPER'].data
            nosta = mp.field('NOSTA')
            noact = mp.field('NOACT')
        except KeyError:
            ag = hdulist['ARRAY_GEOMETRY'].data
            nosta = ag.field('NOSTA')
            noact = ag.field('NOSTA')
        mapper = {}
        for s,a in zip(nosta, noact):
            mapper[s] = a

        # Correct number of visibilities
        self.assertEqual(len(uv.field('FLUX')), data['vis'].shape[0])
        
        # Correct number of frequencies
        for vis in uv.field('FLUX'):
            self.assertEqual(len(vis), 2*len(data['freq']))

        # Correct values
        for bl, vis in zip(uv.field('BASELINE'), uv.field('FLUX')):
            # Convert mapped stands to real stands
            stand1 = mapper[(bl >> 8) & 255]
            stand2 = mapper[bl & 255]

            # Find out which visibility set in the random data corresponds to the 
            # current visibility
            i = 0
            for ant1,ant2 in data['bl']:
                if ant1.stand.id == stand1 and ant2.stand.id == stand2:
                    break
                else:
                    i = i + 1
            
            # Extract the data and run the comparison
            visData = numpy.zeros(len(data['freq']), dtype=numpy.complex64)
            visData.real = vis[0::2]
            visData.imag = vis[1::2]
            for vd, sd in zip(visData, data['vis'][i,:]):
                self.assertAlmostEqual(vd, sd, 8)
            i = i + 1
        
        hdulist.close()

    def test_mapper(self):
        """Test the NOSTA_MAPPER table."""

        testTime = time.time()
        testFile = os.path.join(self.testPath, 'idi-test-SM.fits')
        
        # Get some data
        data = self.__initData()
        
        # Start the file
        fits = fitsidi.IDI(testFile, refTime=testTime)
        fits.setStokes(['xx'])
        fits.setFrequency(data['freq'])
        fits.setGeometry(data['site'], data['antennas'])
        fits.addDataSet(testTime, 6.0, data['bl'], data['vis'])
        fits.write()

        # Open the file and examine
        hdulist = pyfits.open(testFile)
        extNames = [hdu.name for hdu in hdulist]
        maxStand = -1
        for ant in data['antennas']:
            if ant.stand.id > maxStand:
                maxStand = ant.stand.id
        if maxStand > 255:
            self.assertTrue('NOSTA_MAPPER' in extNames)

            # Make sure the mapper makes sense
            mp = hdulist['NOSTA_MAPPER'].data
            ag = hdulist['ARRAY_GEOMETRY'].data
            mNoSta = mp.field('NOSTA')
            aNoSta = ag.field('NOSTA')
            mNoAct = mp.field('NOACT')
            aAnNam = ag.field('ANNAME')
            for msta, mact, asta, anam in zip(mNoSta, mNoAct, aNoSta, aAnNam):
                self.assertEqual(msta, asta)
                self.assertEqual(mact, int(anam[3:]))

        hdulist.close()
        
    def test_multi_if(self):
        """Test writing more than one IF to a FITS IDI file."""
        
        testTime = time.time()
        testFile = os.path.join(self.testPath, 'idi-test-MultiIF.fits')
        
        # Get some data
        data = self.__initData()
        
        # Start the file
        fits = fitsidi.IDI(testFile, refTime=testTime)
        fits.setStokes(['xx'])
        fits.setFrequency(data['freq'])
        fits.setFrequency(data['freq']+10e6)
        fits.setGeometry(data['site'], data['antennas'])
        fits.addDataSet(testTime, 6.0, data['bl'], 
                        numpy.concatenate([data['vis'], 10*data['vis']], axis=1))
        fits.write()

        # Open the file and examine
        hdulist = pyfits.open(testFile)
        # Check that all of the extensions are there
        extNames = [hdu.name for hdu in hdulist]
        for ext in ['ARRAY_GEOMETRY', 'FREQUENCY', 'ANTENNA', 'BANDPASS', 'SOURCE', 'UV_DATA']:
            self.assertTrue(ext in extNames)
            
        # Load the mapper
        try:
            mp = hdulist['NOSTA_MAPPER'].data
            nosta = mp.field('NOSTA')
            noact = mp.field('NOACT')
        except KeyError:
            ag = hdulist['ARRAY_GEOMETRY'].data
            nosta = ag.field('NOSTA')
            noact = ag.field('NOSTA')
        mapper = {}
        for s,a in zip(nosta, noact):
            mapper[s] = a
            
        # Correct number of visibilities
        uv = hdulist['UV_DATA'].data
        self.assertEqual(len(uv.field('FLUX')), data['vis'].shape[0])
        
        # Correct number of frequencies
        for vis in uv.field('FLUX'):
            self.assertEqual(len(vis), 2*2*len(data['freq']))

        # Correct values
        for bl, vis in zip(uv.field('BASELINE'), uv.field('FLUX')):
            # Convert mapped stands to real stands
            stand1 = mapper[(bl >> 8) & 255]
            stand2 = mapper[bl & 255]

            # Find out which visibility set in the random data corresponds to the 
            # current visibility
            i = 0
            for ant1,ant2 in data['bl']:
                if ant1.stand.id == stand1 and ant2.stand.id == stand2:
                    break
                else:
                    i = i + 1
            
            # Extract the data and run the comparison - IF 1
            visData = numpy.zeros(2*len(data['freq']), dtype=numpy.complex64)
            visData.real = vis[0::2]
            visData.imag = vis[1::2]
            for vd, sd in zip(visData[:len(data['freq'])], data['vis'][i,:]):
                self.assertAlmostEqual(vd, sd, 8)
                
            # Extract the data and run the comparison - IF 2
            visData = numpy.zeros(2*len(data['freq']), dtype=numpy.complex64)
            visData.real = vis[0::2]
            visData.imag = vis[1::2]
            for vd, sd in zip(visData[len(data['freq']):], 10*data['vis'][i,:]):
                self.assertAlmostEqual(vd, sd, 8)
                
        hdulist.close()
        
    def tearDown(self):
        """Remove the test path directory and its contents"""

        tempFiles = os.listdir(self.testPath)
        for tempFile in tempFiles:
            os.unlink(os.path.join(self.testPath, tempFile))
        os.rmdir(self.testPath)
        self.testPath = None


class aipsidi_tests(unittest.TestCase):
    """A unittest.TestCase collection of unit tests for the lsl.writer.fitsidi.aips
    class."""

    testPath = None

    def setUp(self):
        """Turn off all numpy warnings and create the temporary file directory."""

        numpy.seterr(all='ignore')
        self.testPath = tempfile.mkdtemp(prefix='test-aipsidi-', suffix='.tmp')

    def __initData(self):
        """Private function to generate a random set of data for writing a FITS
        IDI file.  The data is returned as a dictionary with keys:
        * freq - frequency array in Hz
        * site - lwa.common.stations object
        * stands - array of stand numbers
        * bl - list of baseline pairs in real stand numbers
        * vis - array of visibility data in baseline x freq format
        """

        # Frequency range
        freq = numpy.arange(0,512)*20e6/512 + 40e6
        # Site and stands
        site = lwa_common.lwa1
        antennas = site.getAntennas()[0:40:2]
        
        # Set baselines and data
        blList = uvUtils.getBaselines(antennas, IncludeAuto=True, Indicies=False)
        visData = numpy.random.rand(len(blList), len(freq))
        visData = visData.astype(numpy.complex64)

        return {'freq': freq, 'site': site, 'antennas': antennas, 'bl': blList, 'vis': visData}

    def test_write_tables(self):
        """Test if the AIPS IDI writer writes all of the tables."""

        testTime = time.time()
        testFile = os.path.join(self.testPath, 'idi-test-W.fits')
        
        # Get some data
        data = self.__initData()
        
        # Start the file
        fits = fitsidi.AIPS(testFile, refTime=testTime)
        fits.setStokes(['xx'])
        fits.setFrequency(data['freq'])
        fits.setGeometry(data['site'], data['antennas'])
        fits.addComment('This is a comment')
        fits.addHistory('This is history')
        fits.addDataSet(testTime, 6.0, data['bl'], data['vis'])
        fits.write()

        # Open the file and examine
        hdulist = pyfits.open(testFile)
        # Check that all of the extensions are there
        extNames = [hdu.name for hdu in hdulist]
        for ext in ['ARRAY_GEOMETRY', 'FREQUENCY', 'ANTENNA', 'BANDPASS', 'SOURCE', 'UV_DATA']:
            self.assertTrue(ext in extNames)
        # Check the comments and history
        self.assertTrue('This is a comment' in str(hdulist[0].header['COMMENT']).split('\n'))
        self.assertTrue('This is history' in str(hdulist[0].header['HISTORY']).split('\n'))
        
        hdulist.close()
        
    def test_array_geometry(self):
        """Test the AIPS IDI ARRAY_GEOMETRY table."""

        testTime = time.time()
        testFile = os.path.join(self.testPath, 'idi-test-AG.fits')
        
        # Get some data
        data = self.__initData()
        
        # Start the file
        fits = fitsidi.AIPS(testFile, refTime=testTime)
        fits.setStokes(['xx'])
        fits.setFrequency(data['freq'])
        fits.setGeometry(data['site'], data['antennas'])
        fits.addDataSet(testTime, 6.0, data['bl'], data['vis'])
        fits.write()

        # Open the file and examine
        hdulist = pyfits.open(testFile)
        ag = hdulist['ARRAY_GEOMETRY'].data
        # Correct number of stands
        self.assertEqual(len(data['antennas']), len(ag.field('NOSTA')))

        # Correct stand names
        names = ['L%03i' % ant.stand.id for ant in data['antennas']]
        for name, anname in zip(names, ag.field('ANNAME')):
            self.assertEqual(name, anname)

        hdulist.close()

    def test_frequency(self):
        """Test the AIPS IDI FREQUENCY table."""

        testTime = time.time()
        testFile = os.path.join(self.testPath, 'idi-test-FQ.fits')
        
        # Get some data
        data = self.__initData()
        
        # Start the file
        fits = fitsidi.AIPS(testFile, refTime=testTime)
        fits.setStokes(['xx'])
        fits.setFrequency(data['freq'])
        fits.setGeometry(data['site'], data['antennas'])
        fits.addDataSet(testTime, 6.0, data['bl'], data['vis'])
        fits.write()

        # Open the file and examine
        hdulist = pyfits.open(testFile)
        fq = hdulist['FREQUENCY'].data
        # Correct number of FREQIDs
        self.assertEqual(len(fq.field('FREQID')), 1)

        # Correct channel width
        self.assertAlmostEqual(fq.field('CH_WIDTH')[0], numpy.abs(data['freq'][1]-data['freq'][0]), 4)

        # Correct bandwidth
        self.assertAlmostEqual(fq.field('TOTAL_BANDWIDTH')[0], numpy.abs(data['freq'][-1]-data['freq'][0]).astype(numpy.float32), 4)

        # Correct sideband
        self.assertEqual(fq.field('SIDEBAND')[0], 1)

        hdulist.close()

    def test_antenna(self):
        """Test the AIPS IDI ANTENNA table."""

        testTime = time.time()
        testFile = os.path.join(self.testPath, 'idi-test-AN.fits')
        
        # Get some data
        data = self.__initData()
        
        # Start the file
        fits = fitsidi.AIPS(testFile, refTime=testTime)
        fits.setStokes(['xx'])
        fits.setFrequency(data['freq'])
        fits.setGeometry(data['site'], data['antennas'])
        fits.addDataSet(testTime, 6.0, data['bl'], data['vis'])
        fits.write()

        # Open the file and examine
        hdulist = pyfits.open(testFile)
        an = hdulist['ANTENNA'].data
        # Correct number of stands
        self.assertEqual(len(data['antennas']), len(an.field('ANTENNA_NO')))

        # Correct FREQIDs
        for freqid in an.field('FREQID'):
            self.assertEqual(freqid, 1)

        hdulist.close()

    def test_bandpass(self):
        """Test the AIPS IDI BANDPASS table."""

        testTime = time.time()
        testFile = os.path.join(self.testPath, 'idi-test-BP.fits')
        
        # Get some data
        data = self.__initData()
        
        # Start the file
        fits = fitsidi.AIPS(testFile, refTime=testTime)
        fits.setStokes(['xx'])
        fits.setFrequency(data['freq'])
        fits.setGeometry(data['site'], data['antennas'])
        fits.addDataSet(testTime, 6.0, data['bl'], data['vis'])
        fits.write()

        # Open the file and examine
        hdulist = pyfits.open(testFile)
        bp = hdulist['BANDPASS'].data
        # Correct number of entries
        self.assertEqual(len(data['antennas']), len(bp.field('ANTENNA_NO')))

        # Correct Source ID number
        for src in bp.field('SOURCE_ID'):
            self.assertEqual(src, 0)

        # Correct FREQIDs
        for freqid in bp.field('FREQID'):
            self.assertEqual(freqid, 1)

        hdulist.close()

    def test_source(self):
        """Test the AIPS IDI SOURCE table."""

        testTime = time.time()
        testFile = os.path.join(self.testPath, 'idi-test-SO.fits')
        
        # Get some data
        data = self.__initData()
        
        # Start the file
        fits = fitsidi.AIPS(testFile, refTime=testTime)
        fits.setStokes(['xx'])
        fits.setFrequency(data['freq'])
        fits.setGeometry(data['site'], data['antennas'])
        fits.addDataSet(testTime, 6.0, data['bl'], data['vis'])
        fits.write()

        # Open the file and examine
        hdulist = pyfits.open(testFile)
        so = hdulist['SOURCE'].data
        # Correct number of entries
        self.assertEqual(len(so.field('SOURCE_ID')), 1)

        # Correct Source ID number
        self.assertEqual(so.field('SOURCE_ID'), 1)

        hdulist.close()

    def test_uvdata(self):
        """Test the AIPS IDI UV_DATA table."""

        testTime = time.time()
        testFile = os.path.join(self.testPath, 'idi-test-UV.fits')
        
        # Get some data
        data = self.__initData()
        
        # Start the file
        fits = fitsidi.AIPS(testFile, refTime=testTime)
        fits.setStokes(['xx'])
        fits.setFrequency(data['freq'])
        fits.setGeometry(data['site'], data['antennas'])
        fits.addDataSet(testTime, 6.0, data['bl'], data['vis'])
        fits.write()

        # Open the file and examine
        hdulist = pyfits.open(testFile)
        uv = hdulist['UV_DATA'].data

        # Load the mapper
        try:
            mp = hdulist['NOSTA_MAPPER'].data
            nosta = mp.field('NOSTA')
            noact = mp.field('NOACT')
        except KeyError:
            ag = hdulist['ARRAY_GEOMETRY'].data
            nosta = ag.field('NOSTA')
            noact = ag.field('NOSTA')
        mapper = {}
        for s,a in zip(nosta, noact):
            mapper[s] = a

        # Correct number of visibilities
        self.assertEqual(len(uv.field('FLUX')), data['vis'].shape[0])
        
        # Correct number of frequencies
        for vis in uv.field('FLUX'):
            self.assertEqual(len(vis), 2*len(data['freq']))

        # Correct values
        for bl, vis in zip(uv.field('BASELINE'), uv.field('FLUX')):
            # Convert mapped stands to real stands
            stand1 = mapper[(bl >> 8) & 255]
            stand2 = mapper[bl & 255]

            # Find out which visibility set in the random data corresponds to the 
            # current visibility
            i = 0
            for ant1,ant2 in data['bl']:
                if ant1.stand.id == stand1 and ant2.stand.id == stand2:
                    break
                else:
                    i = i + 1
            
            # Extract the data and run the comparison
            visData = numpy.zeros(len(data['freq']), dtype=numpy.complex64)
            visData.real = vis[0::2]
            visData.imag = vis[1::2]
            for vd, sd in zip(visData, data['vis'][i,:]):
                self.assertAlmostEqual(vd, sd, 8)
            i = i + 1
        
        hdulist.close()

    def test_mapper(self):
        """Test the AIPS IDI NOSTA_MAPPER table."""

        testTime = time.time()
        testFile = os.path.join(self.testPath, 'idi-test-SM.fits')
        
        # Get some data
        data = self.__initData()
        
        # Start the file
        fits = fitsidi.AIPS(testFile, refTime=testTime)
        fits.setStokes(['xx'])
        fits.setFrequency(data['freq'])
        fits.setGeometry(data['site'], data['antennas'])
        fits.addDataSet(testTime, 6.0, data['bl'], data['vis'])
        fits.write()

        # Open the file and examine
        hdulist = pyfits.open(testFile)
        extNames = [hdu.name for hdu in hdulist]
        maxStand = -1
        for ant in data['antennas']:
            if ant.stand.id > maxStand:
                maxStand = ant.stand.id
        if maxStand > 99:
            self.assertTrue('NOSTA_MAPPER' in extNames)

            # Make sure the mapper makes sense
            mp = hdulist['NOSTA_MAPPER'].data
            ag = hdulist['ARRAY_GEOMETRY'].data
            mNoSta = mp.field('NOSTA')
            aNoSta = ag.field('NOSTA')
            mNoAct = mp.field('NOACT')
            aAnNam = ag.field('ANNAME')
            for msta, mact, asta, anam in zip(mNoSta, mNoAct, aNoSta, aAnNam):
                self.assertEqual(msta, asta)
                self.assertEqual(mact, int(anam[1:]))

        hdulist.close()

    def tearDown(self):
        """Remove the test path directory and its contents"""

        tempFiles = os.listdir(self.testPath)
        for tempFile in tempFiles:
            os.unlink(os.path.join(self.testPath, tempFile))
        os.rmdir(self.testPath)
        self.testPath = None


class fitsidi_test_suite(unittest.TestSuite):
    """A unittest.TestSuite class which contains all of the lsl.reader units 
    tests."""
    
    def __init__(self):
        unittest.TestSuite.__init__(self)
        
        loader = unittest.TestLoader()
        self.addTests(loader.loadTestsFromTestCase(fitsidi_tests))
        self.addTests(loader.loadTestsFromTestCase(aipsidi_tests))


if __name__ == '__main__':
    unittest.main()
